#! /bin/sh

# TX6.SH -- A Bourne clone of TX6
# Copyright (C) Nils M Holm, 2000. All rights reserved.
# See the file LICENSE for conditions of use
#
# This is an attempt to simulate the TX6 compiler driver in a shell script.
# Not all options of the original version are supported.
# The T3X-Version cannot be used in the generic port, because no SYSTEM
# interface exists.

VERSION="T3X Release 6.8"

if [ "$T3XDIR" = "" ]; then
	T3XDIR=/usr/local/t3x/r6
fi

BINDIR=$T3XDIR/bin
CLASSDIR=$T3XDIR/classes
BINDIR=../compiler
CLASSDIR=../classes

TXOPT_DFLT="$BINDIR/txx $BINDIR/txopt"
TXDPF_DFLT="$BINDIR/txx $BINDIR/txdpf"

TXTRN="$BINDIR/txx $BINDIR/txtrn"
TXLD="$BINDIR/txx $BINDIR/txld"
TXOPT="$TXOPT_DFLT"
TXDPF="$TXDPF_DFLT"

VERBOSE=0
PRESERVE=0
LOAD=1
R5COMPAT=0
BANGHDR=NA
MAKECLASS=0
OUTFILE=NA
LONGUSAGE=0
LICENSE=0

USAGE="Usage: tx6 [-?h] [-B dir] [-C dir] [-o file] [-5DOcepv] file"

while true; do
	case "x$1" in
		x-\?|x-h)
			LONGUSAGE=1
			break
			;;
		x-B)	if [ "$2" != "" ]; then
				shift
			fi
			BINDIR=$1
			;;
		x-C)	if [ "$2" != "" ]; then
				shift
			fi
			CLASSDIR=$1
			;;
		x-o)	if [ "$2" != "" ]; then
				shift
			fi
			OUTFILE=$1
			;;
		x-O0)	TXOPT=NA
			;;
		x-D0)	TXDPF=NA
			;;
		x-O|x-O1)
			TXOPT=$TXOPT_DFLT
			;;
		x-D|x-D1)
			TXDPF=$TXDPF_DFLT
			;;
		x-V)	LICENSE=1
			break
			;;
		x-5)	R5COMPAT=1
			;;
		x-c)	LOAD=0
			;;
		x-e)	BANGHDR="#! $BINDIR/txx"
			;;
		x-p)	PRESERVE=1
			;;
		x-v)	VERBOSE=1
			echo "$VERSION"
			;;
		x-vv)	VERBOSE=2
			echo "$VERSION"
			;;
		x-*|x)
			echo $USAGE
			exit 1
			;;
		x*)	break
			;;
	esac
	shift
done

if [ $LONGUSAGE = 1 ]; then
echo
echo $USAGE
cat <<EOT

-B dir  search binaries in 'dir'
-C dir  search classes in 'dir'
-D{01}  enable/disable dead procedure filter
-O{01}  enable/disable tcode optimization
-V      display the conditions of use
-c      compile only (don't load)
-e      create a #! header (Unix only)
-h      display this summary (also -?)
-o file direct output to 'file'
-p      preserve CLASSDIR
-v      verbose operation (more v's = more output)
-5      enable T3X-R5 compatibility mode

EOT
exit 0
fi

if [ $LICENSE = 1 ]; then
echo
cat <<EOT
T3X -- A Compiler for the Minimum Procedural Language T3X
Copyright (C) 1996-2000 Nils M Holm.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
EOT
echo
exit 0
fi

FILE=$1
OFILE=`basename $FILE .t`
TMP=tx$$.tmp

if [ ! -f $FILE ]; then
	echo "TX6: $FILE: no such file"
	exit 1
fi

if [ ! -f $OFILE.t ]; then
	echo "TX6: $FILE: file has no .t suffix"
	exit 1
fi

if [ "$OUTFILE" != "NA" ]; then
	OFILE=$OUTFILE
fi

if [ $PRESERVE = 0 ]; then
	rm -f CLASSDIR
fi

if [ $VERBOSE != 0 ]; then
	echo $TXTRN: $FILE '->' $OFILE
fi
$TXTRN <$FILE >$OFILE
if [ -f T3X.ERR -o $? != 0 ]; then
	rm -f T3X.ERR $TMP $OFILE
	echo "TX6: error in compilation. stopped."
	exit 1
fi

if [ $LOAD = 0 ]; then
	exit 0
fi

if [ $R5COMPAT = 1 ]; then
	if [ $VERBOSE = 2 ]; then
		echo "(loading) $CLASSDIR/string.cls"
		echo "(loading) $CLASSDIR/r5c.cls"
	fi
	rm -f $TMP
	cat $CLASSDIR/string.cls >>$TMP
	cat $CLASSDIR/r5c.cls >>$TMP
	cat $OFILE >>$TMP
	mv -f $TMP $OFILE
elif [ -f CLASSDIR ]; then
	rm -f $TMP
	for libname in char iostream memory string; do
		if grep -q \*$libname\* CLASSDIR; then
			if [ $VERBOSE = 2 ]; then
				echo "(loading) $CLASSDIR/$libname.cls"
			fi
			cat $CLASSDIR/$libname.cls >>$TMP
		fi
	done
	cat $OFILE >>$TMP
	mv -f $TMP $OFILE
fi

if [ $VERBOSE != 0 ]; then
	echo $TXLD $FILE '->' $OFILE
fi
mv -f $OFILE $TMP
$TXLD <$TMP >$OFILE
if [ -f T3X.ERR -o $? != 0 ]; then
	rm -f T3X.ERR $TMP $OFILE
	echo "TX6: error in compilation. stopped."
	exit 1
fi

if [ "$TXOPT" != "NA" ]; then
	if [ $VERBOSE != 0 ]; then
		echo $TXOPT $FILE '->' $OFILE
	fi
	mv -f $OFILE $TMP
	$TXOPT <$TMP >$OFILE
	if [ -f T3X.ERR -o $? != 0 ]; then
		rm -f T3X.ERR $TMP $OFILE
		echo "TX6: error in compilation. stopped."
		exit 1
	fi
fi

if [ "$TXDPF" != "NA" ]; then
	if [ $VERBOSE != 0 ]; then
		echo $TXDPF $FILE '->' $OFILE
	fi
	mv -f $OFILE $TMP
	$TXDPF <$TMP >$OFILE
	if [ -f T3X.ERR -o $? != 0 ]; then
		rm -f T3X.ERR $TMP $OFILE
		echo "TX6: error in compilation. stopped."
		exit 1
	fi
fi

if [ "$BANGHDR" != "NA" ]; then
	mv -f $OFILE $TMP
	echo "$BANGHDR" >$OFILE
	cat $TMP >>$OFILE
	chmod +x $OFILE
fi

if [ $PRESERVE = 0 ]; then
	rm -f CLASSDIR
fi
rm -f $TMP
