!	TXLD -- Load Tcode3 programs
!	Copyright (C) 1999,2000 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.
!
!	SYNTAX: CAT MODULE ... | TXLD >OUTFILE

module txld(t3x, tcode);

object	t[t3x];

const	BUFSIZE	=	1025;
const	TEXTLEN	=	257;
const	PUBSIZE	=	6144;	! words
const	EXTSIZE	=	2048;
const	MAXMODS	=	256;

struct	PSYM = P_left, P_right, P_lab;
const	P_name = PSYM;
struct	XMAP = X_id, X_lab;

var	Buffer::BUFSIZE, Cp, Ep;
var	Obuffer::BUFSIZE, Obp;
var	Symbols, Pubtbl[PUBSIZE], Ptop;
var	Exttbl[EXTSIZE], Etop;
var	Modlist[MAXMODS], Mtop;
var	Addr;
var	K64;
var	Base, Top;
var	Errflag;


init() do
	Ep := 1;
	Cp := 1;
	Obp := 0;
	Addr := 0;
	Base := 0;
	Ptop := 0;
	Etop := 0;
	Mtop := 0;
	Symbols := 0;
	K64 := 16384; K64 := K64<<2;	! 32-bit hack
	Errflag := 0;
end


length(s) do var k;
	k := 0;
	while (s::k) k := k+1;
	return k;
end


writep(f, s) t.write(f, s, length(s));


nl(f) do var b::3;
	writep(f, t.newline(b));
end


var	ntoa_buf::32;
ntoa(v) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
        ntoa_buf::31:= 0;
        i := 30;
        while (v \/ i = 30) do
                ntoa_buf::i := v mod 10 + '0'; i := i-1;
                v := v/10;
        end
        if (g) do
                ntoa_buf::i := '-'; i := i-1;
        end
        return @ntoa_buf::(i+1);
end


error(m) do
	writep(T3X.SYSERR, packed"TXLD: ");
	writep(T3X.SYSERR, m);
	writep(T3X.SYSERR, packed" (");
	writep(T3X.SYSERR, ntoa(Addr));
	writep(T3X.SYSERR, packed")");
	nl(T3X.SYSERR);
	t.close(t.open(packed"T3X.ERR", T3X.OWRITE));
	halt 1;
end


rdch() do
	if (Cp >= Ep) do
		Ep := t.read(T3X.SYSIN, Buffer, BUFSIZE-1);
		Cp := 0;
		if (Ep < 0) Ep := 0;
		if (\Ep) return %1;
	end
	Cp := Cp+1;
	Addr := Addr+1;
	return Buffer::(Cp-1);
end


flush() do
	t.write(T3X.SYSOUT, Obuffer, Obp);
	Obp := 0;
end


wrch(n) do
	if (Obp >= BUFSIZE-1) flush();
	Obuffer::Obp := n;
	Obp := Obp+1;
end


extend(x) return x > 32767-> x-K64: x;


report(m, n) do var o;
	writep(T3X.SYSERR, packed"TXLD: ");
	writep(T3X.SYSERR, n);
	writep(T3X.SYSERR, packed": ");
	writep(T3X.SYSERR, m);
	nl(T3X.SYSERR);
	ie (\Errflag)
		t.close(t.open(packed"T3X.ERR", T3X.OWRITE));
	else
		Errflag := 1;
end


findsym(name, pp) do var k, y, r, parent;
	y := Symbols;
	parent := 0;
	k := length(name);
	while (y) do
		r := t.memcomp(name, @y[P_name], k+1);
		if (\r) do
			if (pp) pp[0] := parent->
				@parent[r<0-> P_left: P_right]: 0;
			return y;
		end
		parent := y;
		y := r<0-> y[P_left] : y[P_right];
	end
	if (pp) pp[0] := parent-> @parent[r<0-> P_left: P_right]: 0;
	return 0;
end


addsym(name, l) do
	var	k, y, new, pp;

	y := findsym(name, @pp);
	if (y) do
		report(packed"duplicate symbol", name);
		return y;
	end
	k := length(name);
	if ((Ptop + k + 2) / 2 + PSYM >= PUBSIZE)
		error(packed"symbol table overflow");
	new := @Pubtbl[Ptop];
	Ptop := Ptop + (k+2)/2 + PSYM;
	new[P_lab] := l;
	new[P_left] := 0;
	new[P_right] := 0;
	t.memcopy(@new[P_name], name, k+1);
	ie (pp) do
		pp[0] := new;
	end
	else do
		Symbols := new;
	end
	return new;
end


addref(name, id) do
	var	y, a, new;

	y := findsym(name, 0);
	ie (\y) do
		report(packed"unresolved external", name);
		a := -1;
	end
	else do
		a := y[P_lab];
	end
	if (Etop + XMAP >= EXTSIZE)
		error(packed"link map overflow");
	new := @Exttbl[Etop];
	Etop := Etop+XMAP;
	new[X_id] := id;
	new[X_lab] := a;
	return a;
end


lookup(id) do var i;
	for (i=0, Etop, XMAP) do
		if (Exttbl[i+X_id] = id)
			return Exttbl[i+X_lab];
	end
	report(packed"bad external reference", ntoa(id));
end


addmodule(id) do
	if (Mtop >= MAXMODS) error(packed"too many modules");
	Modlist[Mtop] := id;
	Mtop := Mtop+1;
end


readop() return rdch() | (rdch() << 8);


writeop(n) do
	wrch(n);
	wrch(n>>8);
end


gen1(i, o) do
	wrch(i);
	writeop(o);
end


load() do
	var	op, c, a1, a2, v, i, n, o2;
	var	scan, sbuf[TEXTLEN];

	Top := 0;
	while (1) do
		scan := 0;
		op := rdch();
		if (\Ep) leave;
		if (op & 127 < 1 \/ op & 127 >= tcode.IENDOFSET)
			error(packed"illegal opcode");
		c := op & 128->
			(op = tcode.IINIT \/
			 op = tcode.IINCL \/ op = tcode.IINCI \/
			 op = tcode.IINCG \/
			 op = tcode.IPUB \/ op = tcode.IEXT \/
			 op = tcode.ILSYM \/ op = tcode.IGSYM \/
			 op = tcode.IISYM)-> 2: 1: 0;
		v := op = tcode.ISTR \/ op = tcode.IPSTR \/
			op = tcode.IPUB \/ op = tcode.IEXT \/
			op = tcode.IGSYM \/ op = tcode.ILSYM \/
			op = tcode.IISYM;
		if (c) a1 := readop();
		if (c > 1) a2 := readop();
		if (op = tcode.IINIT) do
			if (a1 \= 3) error(packed"unsupported Tcode version");
			for (i=0, 2) do
				rdch(); readop(); ! swallow CALL,HALT
			end
			Base := Top;
			Etop := 0;
		end
		ie (op = tcode.ICLAB \/ op = tcode.IDLAB \/
			op = tcode.ICREF \/ op = tcode.IDREF \/
			op = tcode.IINCG \/ op = tcode.ILDG \/
			op = tcode.ILDGV \/ op = tcode.ILDLAB \/
			op = tcode.ISAVG \/ op = tcode.ICALL \/
			op = tcode.IBRF \/ op = tcode.IBRT \/
			op = tcode.INBRF \/ op = tcode.INBRT \/
			op = tcode.IJUMP \/ op = tcode.IUNEXT \/
			op = tcode.IDNEXT \/ op = tcode.IPUB \/
			op = tcode.IINIT
		) do
			ie (op = tcode.IINIT) do
				a2 := a2 + Base;
				if (a2 >= Top) Top := a2+1;
			end
			else do
				a1 := a1 + Base;
				if (a1 >= Top) Top := a1+1;
			end
		end
		else if (op = tcode.ICALX) do
			i := lookup(a1);
			if (i \= -1) do
				op := tcode.ICALL;
				a1 := i;
			end
		end
		if (op = tcode.IEXT \/ op = tcode.IPUB) scan := 1;
		if (op = tcode.IINIT) do
			addmodule(a2);
			ie (Base)
				scan := 1;
			else
				a2 := 0;
		end
		if (\scan) do
			wrch(op);
			if (c) writeop(a1);
			if (c > 1) writeop(a2);
		end
		if (v) do
			v := (op = tcode.IPSTR \/ op = tcode.ISTR)-> a1: a2;
			for (i=0, v) do
				n := rdch();
				ie (scan)
					sbuf::i := n;
				else
					wrch(n);
			end
			if (scan) sbuf::i := 0;
		end
		if (Base = 0 /\ op = tcode.IINIT) gen1(tcode.IJUMP, 0);
		if (op = tcode.IPUB) addsym(sbuf, a1);
		if (op = tcode.IEXT /\ addref(sbuf, a1) = -1) do
			wrch(op);
			writeop(a1);
			writeop(v);
			FOR (i=0, v) wrch(sbuf::i);
		end
	end
	gen1(tcode.ICLAB, 0);
	for (i=0, Mtop) if (Modlist[i]) do
		gen1(tcode.ICALL, Modlist[i]);
		gen1(tcode.ICLEAN, 0);
		wrch(tcode.IPOP);
	end
	gen1(tcode.IHALT, 0);
end


do
	init();
	load();
	flush();
end
