!	TXOPT -- A Tcode Optimizer for T3X.
!	Copyright (C) 1999,2000 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.
!
!	SYNTAX: TXOPT <INFILE >OUTFILE

module txopt(t3x, tcode);

object	t[t3x];

const	BUFLEN		= 1024;
const	CODESIZE	= 8192;
const	NLABELS		= 2048;
const	STACKLEN	= 64;
const	XSTACKLEN	= 512;

struct	TCI = tc_op, tc_arg, tc_left, tc_right;	! Tcode instruction

var	Code::CODESIZE;
var	Ctop;
var	Labels[NLABELS], Lbase;
var	Stack[STACKLEN], Sp;
var	Inbuf::BUFLEN, Cp, Ep;
var	Comtab;
var	Xstack[XSTACKLEN*TCI], X0, Xp;


init() do
	Cp := 0;
	Ep := 0;
	Comtab := [
	[ 1, tcode.IDEREF,	0		],
	[ 1, tcode.IDREFB,	tcode.IDREFB	],
	[ 1, tcode.IIND,	0		],
	[ 1, tcode.IINDB,	0		],
	[ 2, tcode.IMUL,	tcode.IMUL	],
	[ 2, tcode.IUMUL,	tcode.IUMUL	],
	[ 2, tcode.IDIV,	0		],
	[ 2, tcode.IUDIV,	0		],
	[ 2, tcode.IMOD,	0		],
	[ 3, tcode.IADD,	tcode.IADD	],
	[ 3, tcode.ISUB,	0		],
	[ 4, tcode.IBAND,	tcode.IBAND	],
	[ 4, tcode.IBOR,	tcode.IBOR	],
	[ 4, tcode.IBXOR,	tcode.IBXOR	],
	[ 4, tcode.IBSHL,	0		],
	[ 4, tcode.IBSHR,	0		],
	[ 5, tcode.IEQU,	tcode.IEQU	],
	[ 5, tcode.INEQU,	tcode.INEQU	],
	[ 6, tcode.ILESS,	tcode.IGRTR	],
	[ 6, tcode.IGRTR,	tcode.ILESS	],
	[ 6, tcode.ILTEQ,	tcode.IGTEQ	],
	[ 6, tcode.IGTEQ,	tcode.ILTEQ	],
	[ 6, tcode.IULESS,	tcode.IUGRTR	],
	[ 6, tcode.IUGRTR,	tcode.IULESS	],
	[ 6, tcode.IULTEQ,	tcode.IULTEQ	],
	[ 6, tcode.IUGTEQ,	tcode.IUGTEQ	],
	[ 0, 0,			0		],
	];
end


length(s) do var k;
	k := 0;
	while (s::k) k := k+1;
	return k;
end


writep(f, s) t.write(f, s, length(s));


nl(f) do var b::3;
	writep(f, t.newline(b));
end


var	ntoa_buf::32;
ntoa(v) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
        ntoa_buf::31:= 0;
        i := 30;
        while (v \/ i = 30) do
                ntoa_buf::i := v mod 10 + '0'; i := i-1;
                v := v/10;
        end
        if (g) do
                ntoa_buf::i := '-'; i := i-1;
        end
        return @ntoa_buf::(i+1);
end


fatal(m, n) do
	writep(T3X.SYSERR, packed"TXOPT: ");
	writep(T3X.SYSERR, m);
	if (n) do
		writep(T3X.SYSERR, ": ");
		writep(T3X.SYSERR, n);
	end
	nl(T3X.SYSERR);
	t.close(t.open(packed"T3X.ERR", T3X.OWRITE));
	writep(T3X.SYSERR, packed"terminating."); nl(T3X.SYSERR);
	halt 1;
end


getw(a) do
	var	n, k64;

	k64 := 16384; k64 := k64 << 2;
	n := Code::(a+1) << 8 | Code::a;
	return (n > 32767)-> n-k64: n;	! 32-bit hack !
end


putw(n, v) do
	Code::n := v & 255;
	Code::(n+1) := v >> 8 & 255;
end


next(p) do var i;
	i := Code::p;
	p := p+1;
	if (i & 128) do
		p := p+2;
                if (    i = tcode.IINCL \/ i = tcode.IINCG \/
                	i = tcode.IINCI \/
                        i = tcode.IINIT \/
                        i = tcode.IPUB \/ i = tcode.IEXT \/
                        i = tcode.ILSYM \/ i = tcode.IGSYM \/ i = tcode.IISYM
                )
			p := p+2;
                if (    i = tcode.ISTR \/ i = tcode.IPSTR \/
                        i = tcode.IPUB \/ i = tcode.IEXT \/
                        i = tcode.IGSYM \/ i = tcode.ILSYM \/ i = tcode.IISYM
                )
			p := p+getw(p-2);
	end
	return p;
end


rdch() do
	if (Cp >= Ep) do
		if (Ep = -1) return -1;
		Ep := t.read(T3X.SYSIN, Inbuf, BUFLEN);
		if (Ep < 1) do
			Ep := -1;
			return -1;
		end
		Cp := 0;
	end
	Cp := Cp+1;
	return Inbuf::(Cp-1);
end


copyi(i) do var k, j;
	Code::Ctop := i;
	Ctop := Ctop+1;
	if (Ctop >= CODESIZE-4) fatal(packed"procedure too big", 0);
	if (i & 128) do
		k := rdch() | (rdch()<<8);
		Code::Ctop := k;
		Code::(Ctop+1) := k>>8;
		Ctop := Ctop+2;
		if (i = tcode.ICLAB /\ k < Lbase) Lbase := k;
		if (    i = tcode.IINCL \/ i = tcode.IINCG \/
			i = tcode.IINCI \/
			i = tcode.IINIT \/
			i = tcode.IPUB \/ i = tcode.IEXT \/
			i = tcode.ILSYM \/ i = tcode.IGSYM \/ i = tcode.IISYM
		) do
			k := rdch() | (rdch()<<8);
			Code::Ctop := k;
			Code::(Ctop+1) := k>>8;
			Ctop := Ctop+2;
		end
		if (    i = tcode.ISTR \/ i = tcode.IPSTR \/
			i = tcode.IPUB \/ i = tcode.IEXT \/
			i = tcode.IGSYM \/ i = tcode.ILSYM \/ i = tcode.IISYM
		) do
			for (j=0, k) do
				Code::(Ctop+j) := rdch();
			end
			Ctop := Ctop+k;
		end
	end
	return rdch();
end


readproc() do var op, k;
	Ctop := 0;
	Lbase := 32767;
	op := rdch();
	while (op \= tcode.IHDR /\ op \= -1) op := copyi(op);
	if (op = -1) return -1;
	while (op \= tcode.IEND /\ op \= -1) op := copyi(op);
	if (op = -1) fatal(packed"unexpected end of procedure", 0);
	Code::Ctop := tcode.IEND;
	Ctop := Ctop+1;
end


nop3(p) do
	Code::p := tcode.INOP;
	Code::(p+1) := tcode.INOP;
	Code::(p+2) := tcode.INOP;
end


fold_mul(p, a1, a2, t1, t2, v1, v2) do
	if (	t1 = tcode.INUM /\ v1 = 0 \/		! x*0, 0*x --> 0
		t2 = tcode.INUM /\ v2 = 0
	) do
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = 1) do		! x*1 --> x
		nop3(a1);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! 1*x --> x
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = %1) do		! x*-1 --> -x
		nop3(a1);
		Code::p := tcode.INEG;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = %1) do		! -1*x --> -x
		nop3(a2);
		Code::p := tcode.INEG;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_div(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0/x --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x/0 --> _|_
		fatal(packed"constant divide by zero", 0);
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! x/1 --> x
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = %1) do		! x/-1 --> -x
		nop3(a2);
		Code::p := tcode.INEG;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_mod(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0 mod x --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x mod 0 --> _|_
		fatal(packed"constant divide by zero", 0);
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! x mod 1 --> 0
		nop3(a2);
		Code::p := tcode.INOP;
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_add(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0+x --> x
		nop3(a1);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x+0 --> x
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_sub(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0-x --> -x
		nop3(a1);
		Code::p := tcode.INEG;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x-0 --> x
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_band(p, a1, a2, t1, t2, v1, v2) do
	if (	t1 = tcode.INUM /\ v1 = 0 \/
		t2 = tcode.INUM /\ v2 = 0
	) do						! 0&x, x&0 --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = %1) do		! %1&x --> x
		nop3(a1);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = %1) do		! x&%1 --> x
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_bor(p, a1, a2, t1, t2, v1, v2) do
	if (	t1 = tcode.INUM /\ v1 = %1 \/
		t2 = tcode.INUM /\ v2 =%1 
	) do						! %1|x, x|%1 --> %1
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, %1);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0|x --> x
		nop3(a1);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x|0 --> x
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_bshl(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0<<x --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x<<0 --> x
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_bshr(p, a1, a2, t1, t2, v1, v2)
	return fold_bshl(p, a1, a2, t1, t2, v1, v2);
	

fold_bxor(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = %1) do		! x^%1 --> ~x
		nop3(a1);
		Code::p := tcode.IBNOT;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = %1) do		! x^%1 --> ~x
		nop3(a2);
		Code::p := tcode.IBNOT;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_deref(p, a1, a2, t1, t2, v1, v2) do
	! Fold the address computation of local vector members
	! at constant offsets
	if (t1 = tcode.ILDLV /\ t2 = tcode.INUM) do	! x[N] --> @(x-N)
		putw(a1+1, v1 - v2);
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x[0] --> x[]
		nop3(a2);
		Code::p := tcode.INOP;
		return %1;
	end
	return 0;
end


fold_umul(p, a1, a2, t1, t2, v1, v2) do
	if (	t1 = tcode.INUM /\ v1 = 0 \/		! x.*0, 0.*x --> 0
		t2 = tcode.INUM /\ v2 = 0
	) do
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = 1) do		! x.*1 --> x
		nop3(a1);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! 1.*x --> x
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_udiv(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0./x --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x./0 --> _|_
		fatal(packed"constant divide by zero", 0);
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! x./1 --> x
		nop3(a2);
		Code::p := tcode.INOP;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end


fold_binary(p) do
	var	op, t1, t2, a1, a2, v1, v2;

	op := Code::p;
	a1 := Stack[Sp-2];
	a2 := Stack[Sp-1];
	t1 := Code::a1;
	t2 := Code::a2;
	v1 := getw(a1+1);
	v2 := getw(a2+1);
	Sp := Sp-1;
	! Fold operations with two constant operands
	! constant OP constant --> constant
	if (t1 = tcode.INUM /\ t2 = tcode.INUM) do
		v1 := getw(a1+1);
		v2 := getw(a2+1);
		if (\v2 /\ (op = tcode.IMOD \/ op = tcode.IDIV))
			fatal(packed"constant divide by zero", 0);
		ie (op = tcode.IDEREF) v1 := v1 - v2;
		else ie (op = tcode.IMUL) v1 := v1 * v2;
		else ie (op = tcode.IDIV) v1 := v1 / v2;
		else ie (op = tcode.IMOD) v1 := v1 mod v2;
		else ie (op = tcode.IADD) v1 := v1 + v2;
		else ie (op = tcode.ISUB) v1 := v1 - v2;
		else ie (op = tcode.IBAND) v1 := v1 & v2;
		else ie (op = tcode.IBOR) v1 := v1 | v2;
		else ie (op = tcode.IBXOR) v1 := v1 ^ v2;
		else ie (op = tcode.IBSHL) v1 := v1 << v2;
		else ie (op = tcode.IBSHR) v1 := v1 >> v2;
		else ie (op = tcode.IEQU) v1 := v1 = v2;
		else ie (op = tcode.INEQU) v1 := v1 \= v2;
		else ie (op = tcode.ILESS) v1 := v1 < v2;
		else ie (op = tcode.IGRTR) v1 := v1 > v2;
		else ie (op = tcode.ILTEQ) v1 := v1 <= v2;
		else ie (op = tcode.IGTEQ) v1 := v1 >= v2;
		else ie (op = tcode.IUMUL) v1 := v1 .* v2;
		else ie (op = tcode.IUDIV) v1 := v1 ./ v2;
		else ie (op = tcode.IULESS) v1 := v1 .< v2;
		else ie (op = tcode.IUGRTR) v1 := v1 .> v2;
		else ie (op = tcode.IULTEQ) v1 := v1 .<= v2;
		else if (op = tcode.IUGTEQ) v1 := v1 .>= v2;
		putw(a1+1, v1);
		nop3(a2);
		Code::p := tcode.INOP;
		return 0;
	end
	if (op = tcode.IDEREF /\ fold_deref(p, a1, a2, t1, t2, v1, v2))
		return 0;
	if (op = tcode.IMUL /\ fold_mul(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IDIV /\ fold_div(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IMOD /\ fold_mod(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IADD /\ fold_add(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.ISUB /\ fold_sub(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IBAND /\ fold_band(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IBOR /\ fold_bor(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IBSHL /\ fold_bshl(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IBSHR /\ fold_bshl(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IBXOR /\ fold_bxor(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IUMUL /\ fold_umul(p, a1, a2, t1, t2, v1, v2)) return 0;
	if (op = tcode.IUDIV /\ fold_udiv(p, a1, a2, t1, t2, v1, v2)) return 0;
	! otherwise, make sure the result is non-constant
	Stack[Sp-1] := CODESIZE-1;
end


fold_inc(p, start) do		! x := x+N --> inc(x,N)
	var	op, a;		! x := x-N --> inc(x,-N)
	var	q, q2;

	op := Code::p;
	if (	Code::start = tcode.ILDL /\ op = tcode.ISAVL \/
		Code::start = tcode.ILDG /\ op = tcode.ISAVG \/
		Code::start = tcode.ILDI /\ op = tcode.ISAVI
	) do
		a := getw(start+1);
		if (a \= getw(p+1)) return 0;
		q := next(start);
		if (Code::q \= tcode.INUM) return 0;
		q2 := next(q);
		if (Code::q2 \= tcode.IADD /\ Code::q2 \= tcode.ISUB) return 0;
		Code::start := Code::start = tcode.ILDL-> tcode.IINCL:
				Code::start = tcode.ILDI-> tcode.IINCI:
					tcode.IINCG;
		putw(start+3, Code::q2 = tcode.ISUB-> -getw(start+4):
			getw(start+4));
		nop3(start+5);
		nop3(p);
	end
end


foldexpr() do
	var	i, op, a, v;
	var	state;
	var	last[3];

	Sp := 0;
	i := 0;
	last[0] := 0;
	last[1] := 0;
	last[2] := 0;
	while (i < Ctop) do
		op := Code::i;
		ie (	op = tcode.ILDG \/ op = tcode.ILDGV \/
			op = tcode.ILDL \/ op = tcode.ILDLV \/
			op = tcode.ILDLAB \/ op = tcode.INUM
		) do
			if (Sp+1 >= STACKLEN)
				fatal(packed"FOLDEXPR: stack overflow", 0);
			Stack[Sp] := i;
			Sp := Sp+1;
		end
		else ie (Sp /\ (op = tcode.INEG \/ op = tcode.ILNOT \/
			op = tcode.IBNOT)
		) do
			ie (Code::(Stack[Sp-1]) = tcode.INUM) do
				a := Stack[Sp-1];
				v := getw(a+1);
				ie (op = tcode.INEG) v := -v;
				else ie (op = tcode.ILNOT) v := \v;
				else v := ~v;
				putw(a+1, v);
				Code::i := tcode.INOP;
			end
			else do
				! make sure it is non-constant
				Stack[Sp-1] := CODESIZE-1;
			end
		end
		else ie (Sp>1 /\
			(op = tcode.IDEREF \/ op = tcode.IMUL \/
			op = tcode.IDIV \/ op = tcode.IMOD \/
			op = tcode.IADD \/ op = tcode.ISUB \/
			op = tcode.IBAND \/ op = tcode.IBOR \/
			op = tcode.IBXOR \/ op = tcode.IBSHL \/
			op = tcode.IBSHR \/
			op = tcode.IEQU \/ op = tcode.INEQU \/
			op = tcode.ILESS \/ op = tcode.IGRTR \/
			op = tcode.ILTEQ \/ op = tcode.IGTEQ \/
			op = tcode.IUMUL \/ op = tcode.IUDIV \/
			op = tcode.IULESS \/
			op = tcode.IUGRTR \/ op = tcode.IULTEQ \/
			op = tcode.IUGRTR)
		) do
			fold_binary(i);
		end
		else ie (op = tcode.IDLAB \/ op = tcode.IDECL \/
			op = tcode.IDATA \/ op = tcode.ICREF \/
			op = tcode.IDREF \/ op = tcode.ISTR \/
			op = tcode.IPSTR \/ op = tcode.INOP \/
			op = tcode.ILINE \/ op = tcode.IGSYM \/
			op = tcode.ILSYM \/ op = tcode.IISYM
		) do
			! ignore
		end
		else do
			Sp := 0;
			if (	(op = tcode.ISAVL \/ op = tcode.ISAVG \/
				 op = tcode.ISAVI) /\
				last[0]
			)
				fold_inc(i, last[0]);
		end
		last[0] := last[1];
		last[1] := last[2];
		last[2] := i;
		i := next(i);
	end
end


foldcond() do
	var	i, v, op;
	var	last;

	i := 0;
	last := 0;
	while (i < Ctop) do
		op := Code::i;
		ie (	(op = tcode.IBRF \/ op = tcode.IBRT) /\
			Code::last = tcode.INUM
		) do
			v := getw(last+1);
			! NUM T BRT L --> JUMP L
			! NUM F BRT L --> NOP
			ie (op = tcode.IBRT) do
				nop3(last);
				ie (v)
					Code::i := tcode.IJUMP;
				else
					nop3(i);
			end
			! NUM T BRF L --> NOP
			! NUM F BRF L --> JUMP L
			else do ! BRF
				nop3(last);
				ie (v)
					nop3(i);
				else
					Code::i := tcode.IJUMP;
			end
		end
		else if ((op = tcode.IBRF \/ op = tcode.IBRT) /\
			Code::last = tcode.ILNOT
		) do
			! NOT BRT  --> BRF
			! NOT BRF  --> BRT
			! NOT NBRT --> NBRF
			! NOT NBRF --> NBRT
			Code::i := op = tcode.IBRT-> tcode.IBRF:
				op = tcode.IBRF-> tcode.IBRT:
				op = tcode.INBRT-> tcode.INBRF:
				tcode.INBRT;
			Code::last := tcode.INOP;
		end
		if (Code::i \= tcode.INOP) last := i;
		i := next(i);
	end
end


condense() do
	var	i, j, k, m;

	i := 0;
	j := 0;
	while (i < Ctop) do
		k := next(i) - i;
		if (Code::i \= tcode.INOP) do
			t.memcopy(@Code::j, @Code::i, k);
			j := j+k;
		end
		i := i+k;
	end
	k := Ctop;
	Ctop := j;
	return k-Ctop;
end


isload(x) return x & 127 >= tcode.ILDG & 127 /\ x & 127 <= tcode.ISELF;


isbinop(x) return x >= tcode.IMUL /\ x <= tcode.IUGTEQ \/ x = tcode.IDEREF \/
	x = tcode.IDREFB;


isexpr(x) return
	x = tcode.INEG \/
	x = tcode.ILNOT \/
	x = tcode.IBNOT \/
	x >= tcode.IMUL /\ x <= tcode.IUGTEQ \/	! isbinop(x)
 	isload(x) \/
	x >= tcode.IIND /\ x <= tcode.IDREFB \/	! indirection
	x = tcode.IDUP \/
	x = tcode.ISWAP;


common(a, b) do var op;
	if (a = %1) return b = %1;
	if (b = %1) return a = %1;
	if (\common(Xstack[a+tc_left], Xstack[b+tc_left])) return 0;
	if (\common(Xstack[a+tc_right], Xstack[b+tc_right])) return 0;
	op := Xstack[a+tc_op];
	if (op \= Xstack[b+tc_op]) return 0;
	if (Xstack[a+tc_arg] \= Xstack[b+tc_arg]) return 0;
	! A procedure call in a subscript may have a side-effect!
	if (op = tcode.ICALL \/ op = tcode.ICALR \/ op = tcode.ICALX) return 0;
	return %1;
end


commref(args) do
	var	a0, a1, op;

	a0 := args[0];
	a1 := args[1];
	if (Xstack[a1+tc_left] = %1) return 0;
	op := Xstack[Xstack[a1+tc_left]+tc_op];
	if (op \= tcode.IIND /\ op \= tcode.IINDB) return 0;
	if (\common(a0, Xstack[Xstack[a1+tc_left]+tc_left])) return 0;
	a1 := Xstack[Xstack[a1+tc_left]+tc_left];
	if (a1 = %1) return 0;
	Xstack[a1+tc_op] := tcode.IDUP;
	Xstack[a1+tc_left] := %1;
	Xstack[a1+tc_right] := %1;
end


reorder(p) do var op, base, left, right, tmp;
	if (p = %1) return 0;
	reorder(Xstack[p+tc_left]);
	reorder(Xstack[p+tc_right]);
	op := Xstack[p+tc_op];
	if (\isbinop(op)) return 0;
	for (base=0, tcode.IENDOFSET) do
		if (\Comtab[base][0]) return 0;
		if (op = Comtab[base][1]) do
			if (\Comtab[base][2]) return 0;
			leave;
		end
	end
	op := Xstack[Xstack[p+tc_left]+tc_op];
	for (left=0, tcode.IENDOFSET)
		if (\Comtab[left][0] \/ op = Comtab[left][1])
			leave;
	op := Xstack[Xstack[p+tc_right]+tc_op];
	for (right=0, tcode.IENDOFSET)
		if (\Comtab[right][0] \/ op = Comtab[right][1])
			leave;
	if (\Comtab[right][0] \/ Comtab[left][0])
		return 0;
	if (Comtab[right][0] < Comtab[base][0]) do
		tmp := Xstack[p+tc_left];
		Xstack[p+tc_left] := Xstack[p+tc_right];
		Xstack[p+tc_right] := tmp;
		Xstack[p+tc_op] := Comtab[base][2];
	end
end


writeback(p) do
	if (p = %1) return 0;
	writeback(Xstack[p+tc_left]);
	writeback(Xstack[p+tc_right]);
	if (\Xstack[p+tc_op]) return 0;
	Code::X0 := Xstack[p+tc_op];
	X0 := X0+1;
	if (Xstack[p+tc_op] & 128) do
		Code::X0 := Xstack[p+tc_arg];
		Code::(X0+1) := Xstack[p+tc_arg] >> 8;
		X0 := X0+2;
	end
end


dumptree(p,d) do var i;
	if (p = %1) return 0;
	if (d>1) for (i=0,d-1) writep(T3X.SYSERR, packed".");
	writep(T3X.SYSERR, ntoa(Xstack[p+tc_op]));
	writep(T3X.SYSERR, packed"\s");
	writep(T3X.SYSERR, ntoa(Xstack[p+tc_arg]));
	nl(T3X.SYSERR);
	if (Xstack[p+tc_left] \= %1) writep(T3X.SYSERR, packed"L");
	dumptree(Xstack[p+tc_left], d+1);
	if (Xstack[p+tc_right] \= %1) writep(T3X.SYSERR, packed"R");
	dumptree(Xstack[p+tc_right], d+1);
end


treeopts() do
	var	i, j, n, p, cxedone;
	var	args[XSTACKLEN], ap;
	var	op;

	i := 0;
	X0 := %1;
	cxedone := 0;
	while (i < Ctop) do
		ie (isexpr(Code::i)) do
			if (X0 = %1) do
				X0 := i;
				Xp := 0;
				ap := 0;
			end
			op := Code::i;
			if (Xp >= XSTACKLEN*TCI-TCI) X0 := %1;	! give up
			Xstack[Xp+tc_op] := op;
			ie (op & 128)
				Xstack[Xp+tc_arg] :=
					Code::(i+2)<<8 | Code::(i+1);
			else
				Xstack[Xp+tc_arg] := 0;
			ie (isload(op)) do
				args[ap] := Xp;
				ap := ap+1;
				Xstack[Xp+tc_left] := %1;
				Xstack[Xp+tc_right] := %1;
			end
			else ie (isbinop(op)) do
				ie (ap < 2) do
					X0 := %1;	! give up;
				end
				else do
					Xstack[Xp+tc_left] := args[ap-2];
					Xstack[Xp+tc_right] := args[ap-1];
					ap := ap-1;
					args[ap-1] := Xp;
				end
			end
			else do
				ie (Xp) do
					Xstack[Xp+tc_left] := Xp-TCI;
					Xstack[Xp+tc_right] := %1;
					args[ap-1] := Xp;
				end
				else do
					X0 := %1;	! give up
				end
			end
			Xp := Xp+TCI;
		end
		else do
			if (X0 \= %1) do
				if (ap = 2) commref(args);
				for (j=0, ap) do
					reorder(args[j]);
					writeback(args[j]);
				end
				if (X0 < i) do
					while (X0 < i) do
						Code::X0 := tcode.INOP;
						X0 := X0+1;
					end
					cxedone := 1;
				end
				X0 := %1;
			end
		end
		i := next(i);
	end
	if (cxedone) condense();
end


jumpredir() do
	var	i, j, k;
	var	jrm;

	for (i=0, NLABELS) Labels[i] := %1;
	jrm := 0;
	i := 0;
	while (i < Ctop) do
		if (Code::i = tcode.ICLAB) do
			j := getw(i+1) - Lbase;
			if (j >= NLABELS) fatal(packed"too many labels", 0);
			Labels[j] := i+3;
			if (	i >= 3 /\ Code::(i-3) = tcode.IJUMP /\
				getw(i-2) = j+Lbase
			) do
				nop3(i-3);
				jrm := 1;
			end
		end
		i := next(i);
	end
	i := 0;
	while (i < Ctop) do
		if (Code::i = tcode.IJUMP) do
			j := i;
			while (1) do
				if (Code::j \= tcode.IJUMP) leave;
				k := getw(j+1) - Lbase;
				if (k < 0 \/ k >= NLABELS \/ Labels[k] = %1)
					leave;
					! fatal(packed
					! "JMPREDIR: invalid destination",
					! 0);
				j := Labels[k];
				if (i = j) leave;
			end
			if (i \= j) putw(i+1, k+Lbase);
		end
		i := next(i);
	end
	if (jrm) condense();
end


writeproc() if (Ctop) if (t.write(T3X.SYSOUT, Code, Ctop) \= Ctop)
		fatal(packed"file write error", 0);


do var eof;
	init();
	eof := 0;
	while (\eof) do
		eof := readproc();
		foldexpr();
		foldcond();
		condense();
		treeopts();
		jumpredir();
		writeproc();
	end
end
