!	TXTRN -- a T3XR6->Tcode translator
!	Copyright (C) 1997-2000 Nils M Holm. All rights reserved.
!	See the file LICENSE for conditions of use.
!
!	SYNTAX: TXTRN <T3X-PROGRAM >TCODE-PROGRAM
!
!	TXTRN is a compiler for a simple procedural language
!	called T3X. T3X is a superset of the language T, v3.
!	Release 6 adds object oriented programming support.
!	The translator reads a source program from stdin and
!	writes an equivalent Tcode program to stdout.
!	Additionally, it may create a CLASSDIR file containing
!	the symbol table entries of exported symbols.

module txtrn(t3x, tcode);

object	t[t3x];

!	Setting TUNE16 = 1 speeds up some things on true 16-bit systems
const	TUNE16 = 0;

const	BUFSIZE=	1026,
	OBUFL=		1024,
	SYMBSPACE=	3072,
	CACHESPACE=	4096,
	NSPACE=		8192,
	TEXTLEN=	129,
	MAXTBL=		128;

const	META=	256;

const	CNST=	1,	! Symbol table flags
	GLOBL=	2,
	PROC=	4,
	PROTO=	8,
	SYS=	16,
	PUBLC=	32,
	EXTRN=	64,
	CLSS=	128,
	OBJCT=	256,
	IVAR=	512,
	REDIR=	1024,
	LINKD=	2048;

struct SYMREC = SNAME,
		SVAL,	! CNST: value		other: address
		SSIZE,	! CLSS: elements	OBJCT: class
		SFLAGS;

struct SR2 =	SNL, SNH,
		SVL, SVH,
		SZL, SZH,
		SFL, SFH;

var	Classpaths,	! System class paths [see initvars()]
	Syscp::TEXTLEN,
	Usrcp::TEXTLEN;

var	Symbols[SYMBSPACE], St;
var	Libcache[CACHESPACE+1], Lt;
var	Sym_equ, Sym_mod, Sym_sub, Sym_add, Sym_mul,
	Sym_bar, Sym_not;
var	Names::NSPACE, Nt;
var	Line;
var	File::TEXTLEN;
var	Modname::TEXTLEN, Modctx;
var	Errcount;
var	Token;
var	Text::TEXTLEN;
var	Value;
var	Op;
var	Label, Xlabel;
var	Startlab, Endlab, Exitlab;
var	Looplevl;
var	Locladdr;
var	Buffer::BUFSIZE, Cp, Ep, Lowp, Nomore;
var	Obuf::OBUFL, Obp;
var	Last[5], LL;
var	Ops;
var	Bcslot;
var	Debug, Packstr;
var	Classctx, Classoff;
var	Selfobj;
var	Method_adjust;
var	K64;


! Token definitions

const	ENDFILE=%1;

const	SYMBOL=0, NUMBER=1, STRING=2;

const	BINOP=10, DISOP=11, CONOP=12, UNOP=20, ADDROP=21, LPAREN=30,
	RPAREN=31, LBRACK=40, RBRACK=41, SEMI=50, COMMA=60, ASSIGN=70,
	COND=80, COLON=81, BYTEOP=90, DOT=98, METAOP=99;

struct	OPLEN = OPREC, OLEN, ONAME, OTOK, OCODE;

const	KCALL=100, KCLASS=101, KCONST=102, KDECL=103, KDO=104, KELSE=105,
	KEND=106, KFOR=107, KHALT=108, KIE=109, KIF=110, KIFACE=111,
	KLEAVE=112, KLOOP=113, KMODULE=114, KOBJECT=115, KPACKED=116,
	KPUBLIC=117, KRETURN=118, KSEND=119, KSTRUCT=120, KVAR=121,
	KWHILE=122;

decl	fatal(2), factor(0), expr(0), stmt(0), compound(0), declaration(1);


initvars() do var k;
	Classpaths := [
		packed".",
		packed"",	! reserved for $T3XDIR/classes
		packed"/usr/local/t3x/r6/classes",
		packed"C:/v/t3xr6/classes",
		0,		! empty slot for #classpath
		0
	];
	k := t.getenv(packed"T3XDIR", Syscp, TEXTLEN-9);
	if (k > 0) do
		t.memcopy(@Syscp::k, packed"/classes", 9);
		Classpaths[1] := Syscp;
	end
	St := 0;
	Lt := 0;
	Nt := 0;
	Line := 1;
	File::0 := 0;
	Modname::0 := 0;
	Modctx := %1;
	Errcount := 0;
	Label := 2;
	Xlabel := 1;
	Startlab := 0;
	Endlab := 0;
	Exitlab := 0;
	Looplevl := 0;
	Locladdr := 1;
	Cp := 0;
	Ep := 0;
	Lowp := 0;
	Nomore := 0;
	Obp := 0;
	LL := 0;
	Bcslot := 0;
	Debug := 0;
	Packstr := 0;
	Classctx := %1;
	! Classoff := 0;
	Method_adjust := 0;
	K64 := 16384;	! XXX 32-bit hack
	K64 := K64<<2;
	Ops := [
		[ 6, 1, packed"+",	BINOP,	tcode.IADD	],
		[ 7, 1, packed"*",	BINOP,	tcode.IMUL	],
		[ 0, 1, packed";",	SEMI,	0		],
		[ 0, 1, packed",",	COMMA,	0		],
		[ 0, 1, packed"(",	LPAREN,	0		],
		[ 0, 1, packed")",	RPAREN,	0		],
		[ 0, 1, packed"[",	LBRACK,	0		],
		[ 0, 1, packed"]",	RBRACK,	0		],
		[ 3, 1, packed"=",	BINOP,	tcode.IEQU	],
		[ 5, 1, packed"&",	BINOP,	tcode.IBAND	],
		[ 5, 1, packed"|",	BINOP,	tcode.IBOR	],
		[ 5, 1, packed"^",	BINOP,	tcode.IBXOR	],
		[ 0, 1, packed"@",	ADDROP,	0		],
		[ 0, 1, packed"~",	UNOP,	tcode.IBNOT	],
		[ 0, 1, packed":",	COLON,	0		],
		[ 0, 2, packed"::",	BYTEOP,	0		],
		[ 0, 2, packed":=",	ASSIGN,	0		],
		[ 0, 1, packed"\\",	UNOP,	tcode.ILNOT	],
		[ 1, 2, packed"\\/",	DISOP,	0		],
		[ 3, 2, packed"\\=",	BINOP,	tcode.INEQU	],
		[ 4, 1, packed"<",	BINOP,	tcode.ILESS	],
		[ 5, 2, packed"<<",	BINOP,	tcode.IBSHL	],
		[ 4, 2, packed"<=",	BINOP,	tcode.ILTEQ	],
		[ 4, 1, packed">",	BINOP,	tcode.IGRTR	],
		[ 5, 2, packed">>",	BINOP,	tcode.IBSHR	],
		[ 4, 2, packed">=",	BINOP,	tcode.IGTEQ	],
		[ 6, 1, packed"-",	BINOP,	tcode.ISUB	],
		[ 0, 2, packed"->",	COND,	0		],
		[ 7, 1, packed"/",	BINOP,	tcode.IDIV	],
		[ 2, 2, packed"/\\",	CONOP,	0		],
		[ 0, 1, packed".",	DOT,	0		],
		[ 7, 2, packed"./",	BINOP,	tcode.IUDIV	],
		[ 7, 2, packed".*",	BINOP,	tcode.IUMUL	],
		[ 4, 2, packed".<",	BINOP,	tcode.IULESS	],
		[ 4, 3, packed".<=",	BINOP,	tcode.IULTEQ	],
		[ 4, 2, packed".>",	BINOP,	tcode.IUGRTR	],
		[ 4, 3, packed".>=",	BINOP,	tcode.IUGTEQ	],
		[ 0, 1, packed"#",	METAOP,	0		],
		[ 7, 3, packed"mod",	BINOP,	tcode.IMOD	],
		[ 0, 0, 0,	0,	0			]
	];
end


length(a) do var k;
	k := 0;
	while (a::k) k := k+1;
	return k;
end


strequ(a, b) return \t.memcomp(a, b, length(a)+1);


strcpy(a, b) do var k;
	k := length(b);
	t.memcopy(a, b, k+1);
	return k;
end


writep(f, s) t.write(f, s, length(s));


nl(f) do var b::3;
	t.write(f, t.newline(b), length(b));
end


var	ntoa_buf::32;
ntoa(v, w) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
        ntoa_buf::31:= 0;
        i := 30;
        while (v \/ i = 30) do
                ntoa_buf::i := v mod 10 + '0'; i := i-1;
                v := v/10;
                w := w-1;
        end
        if (g) do
                ntoa_buf::i := '-'; i := i-1;
                w := w-1;
        end
        while (w>0) do
                ntoa_buf::i := '\s'; i := i-1;
                w := w-1;
        end
        return @ntoa_buf::(i+1);
end


!XXX
dumptbl(tbl, lim) do var i, y, err;
	for (i=0, lim, SYMREC) do
		y := @tbl[i];
		err := T3X.SYSERR;
		writep(err, ntoa(y[SVAL], 6));
		writep(err, ntoa(y[SSIZE], 6));
		writep(err, "\s");
		writep(err, y[SNAME]);
		writep(err, packed" (");
		writep(err, y[SFLAGS] & CNST-> packed" CNST": packed"");
		writep(err, y[SFLAGS] & GLOBL-> packed" GLOBL": packed"");
		writep(err, y[SFLAGS] & PROC-> packed" PROC": packed"");
		writep(err, y[SFLAGS] & PROTO-> packed" PROTO": packed"");
		writep(err, y[SFLAGS] & SYS-> packed" SYS": packed"");
		writep(err, y[SFLAGS] & PUBLC-> packed" PUBLC": packed"");
		writep(err, y[SFLAGS] & EXTRN-> packed" EXTRN": packed"");
		writep(err, y[SFLAGS] & CLSS-> packed" CLSS": packed"");
		writep(err, y[SFLAGS] & OBJCT-> packed" OBJCT": packed"");
		writep(err, y[SFLAGS] & IVAR-> packed" IVAR": packed"");
		writep(err, y[SFLAGS] & REDIR-> packed" REDIR": packed"");
		writep(err, packed" )\n");
	end
end


error(m, s) do var o, col;
	col := packed": ";
	writep(T3X.SYSERR, packed"TXTRN: ");
	ie (File::0) do
		writep(T3X.SYSERR, File); writep(T3X.SYSERR, col);
	end
	else if (Modname::0) do
		writep(T3X.SYSERR, Modname); writep(T3X.SYSERR, col);
	end
	writep(T3X.SYSERR, ntoa(Line, 0));
	writep(T3X.SYSERR, col);
	writep(T3X.SYSERR, m);
	if (s) do
		writep(T3X.SYSERR, col);
		writep(T3X.SYSERR, s);
	end
	nl(T3X.SYSERR);
	Errcount := Errcount+1;
	if (Errcount >= 20) do
		Errcount := 0;
		fatal(packed"too many errors", 0);
	end
end


fatal(m, s) do
	error(m, s);
	writep(T3X.SYSERR, packed"terminating."); nl(T3X.SYSERR);
	t.close(t.open(packed"T3X.ERR", T3X.OWRITE));
	halt 1;
end


fillbuf() do var i;
	if (Nomore) return 0;
	for (i=Cp, Ep) Buffer::(i-Cp) := Buffer::i;
	i := Ep-Cp;
	Cp := 0;
	Ep := t.read(T3X.SYSIN, @Buffer::i, BUFSIZE/2-1);
	if (Ep < 1) Nomore := 1;
	Ep := Ep + i;
	Lowp := Ep-TEXTLEN;
end


eof() return Nomore /\ Cp >= Ep;


getce() do var c;
	c := Buffer::Cp; Cp := Cp+1;
	if (c \= '\\') return c;
	c := Buffer::Cp; Cp := Cp+1;
	if (c = 'a' \/ c = 'A') return '\a';
	if (c = 'b' \/ c = 'B') return '\b';
	if (c = 'e' \/ c = 'E') return '\e';
	if (c = 'f' \/ c = 'F') return '\f';
	if (c = 'n' \/ c = 'N') return '\n';
	if (c = 'q' \/ c = 'Q' \/ c = '"') return '"' | META;
	if (c = 'r' \/ c = 'R') return '\r';
	if (c = 's' \/ c = 'S') return '\s';
	if (c = 't' \/ c = 'T') return '\t';
	if (c = 'v' \/ c = 'V') return '\v';
	return c;
end


findkw(s) ie (s::0 = 'c') do
		if (strequ(s, packed"call")) return KCALL;
		if (strequ(s, packed"class")) return KCLASS;
		if (strequ(s, packed"const")) return KCONST;
		return 0;
	end
	else ie (s::0 = 'd') do
		if (strequ(s, packed"decl")) return KDECL;
		if (strequ(s, packed"do")) return KDO;
		return 0;
	end
	else ie (s::0 = 'e') do
		if (strequ(s, packed"else")) return KELSE;
		if (strequ(s, packed"end")) return KEND;
		return 0;
	end
	else ie (s::0 = 'f') do
		if (strequ(s, packed"for")) return KFOR;
		return 0;
	end
	else ie (s::0 = 'h') do
		if (strequ(s, packed"halt")) return KHALT;
		return 0;
	end
	else ie (s::0 = 'i') do
		if (strequ(s, packed"ie")) return KIE;
		if (strequ(s, packed"if")) return KIF;
		if (strequ(s, packed"interface")) return KIFACE;
		return 0;
	end
	else ie (s::0 = 'l') do
		if (strequ(s, packed"leave")) return KLEAVE;
		if (strequ(s, packed"loop")) return KLOOP;
		return 0;
	end
	else ie (s::0 = 'm') do
		if (strequ(s, packed"mod")) return BINOP;
		if (strequ(s, packed"module")) return KMODULE;
		return 0;
	end
	else ie (s::0 = 'o') do
		if (strequ(s, packed"object")) return KOBJECT;
		return 0;
	end
	else ie (s::0 = 'p') do
		if (strequ(s, packed"packed")) return KPACKED;
		if (strequ(s, packed"public")) return KPUBLIC;
		return 0;
	end
	else ie (s::0 = 'r') do
		if (strequ(s, packed"return")) return KRETURN;
		return 0;
	end
	else ie (s::0 = 's') do
		if (strequ(s, packed"send")) return KSEND;
		if (strequ(s, packed"struct")) return KSTRUCT;
		return 0;
	end
	else ie (s::0 = 'v') do
		if (strequ(s, packed"var")) return KVAR;
		return 0;
	end
	else ie (s::0 = 'w') do
		if (strequ(s, packed"while")) return KWHILE;
		return 0;
	end
	else do
		return 0;
	end


findop(c) do var nmatch, i, pos;
	nmatch := 0;
	i := 0;
	pos := %1;
	while (Ops[i][OLEN] > nmatch) do
		if (c = Ops[i][ONAME]::nmatch) do
			c := Buffer::Cp; Cp := Cp+1;
			pos := i;
			nmatch := nmatch+1;
			if (Ops[i][OLEN] > nmatch) do
				pos := %1;
				i := i-1;
			end
		end
		i := i+1;
	end
	Cp := Cp-1;
	Op := pos;
	return pos = %1-> 0: Ops[pos][OTOK];
end


findop2(s) do var i;
	i := 0;
	while (Ops[i][OLEN]) do
		if (strequ(Ops[i][ONAME], s)) return i;
		i := i+1;
	end
	fatal(packed"bad name in findop2()", s);
end


scan() do
	var	c, i, bc::3;
	var	radix;

	if (Cp >= Lowp) fillbuf();
	c := Buffer::Cp; Cp := Cp+1;
	while (1) do
		while (	c = '\s' \/ c = '\t' \/ c = '\n' \/
			c = '\r' \/ c = '\f'
		) do
			if (c = '\n') Line := Line+1;
			c := Buffer::Cp; Cp := Cp+1;
			if (Cp >= Lowp) do
				if (eof()) return ENDFILE;
				fillbuf();
			end
		end
		if (c \= '!') leave;
		while (c \= '\n') do
			if (Cp >= Lowp) do
				if (eof()) return ENDFILE;
				fillbuf();
			end
			c := Buffer::Cp; Cp := Cp+1;
		end
	end
	if (eof()) return ENDFILE;
	if (	'a' <= c /\ c <= 'z' \/
		'A' <= c /\ c <= 'Z' \/
		c = '_'
	) do
		i := 0;
		while (1) do
			ie ('A' <= c /\ c <= 'Z')
				c := c-'A'+'a';
			else if (\('a' <= c /\ c <= 'z' \/
				'0' <= c /\ c <= '9' \/
				c = '_')
			)
				leave;
			if (i >= TEXTLEN-1) fatal(packed"symbol too long", 0);
			Text::i := c;
			i := i+1;
			c := Buffer::Cp; Cp := Cp+1;
		end
		Text::i := 0;
		Cp := Cp-1;
		c := findkw(Text);
		if (c) do
			if (c = BINOP) Op := Sym_mod;
			return c;
		end
		return SYMBOL;
	end
	if ('0' <= c /\ c <= '9' \/ c = '%') do
		Value := 0;
		radix := 10;
		i := 0;
		if (c = '%') do
			i := 1;
			c := Buffer::Cp; Cp := Cp+1;
		end
		if (c = '0') do
			c := Buffer::Cp; Cp := Cp+1;
			if (c = 'x' \/ c = 'X') do
				c := Buffer::Cp; Cp := Cp+1;
				radix := 16;
			end
		end
		while (1) do
			ie (radix = 16 /\ 'A' <= c /\ c <= 'F')
				Value := Value*16 + c-('A'-10);
			else ie (radix = 16 /\ 'a' <= c /\ c <= 'f')
				Value := Value*16 + c-('a'-10);
			else ie ('0' <= c /\ c <= '9')
				Value := Value*radix + c-'0';
			else
				leave;
			c := Buffer::Cp; Cp := Cp+1;
		end
		Cp := Cp-1;
		if (i) Value := -Value;
		return NUMBER;
	end
	if (c = '\'') do
		Value := getce() & ~META;
		c := Buffer::Cp; Cp := Cp+1;
		if (c \= '\'') error(packed"missing '\\''", 0);
		return NUMBER;
	end
	if (c = '"') do
		i := 0;
		c := getce();
		while (c \= '"') do
			if (i >= TEXTLEN-2) fatal(packed"string too long", 0);
			Text::i := c & ~META;
			i := i+1;
			c := getce();
			if (eof()) fatal(packed"unexpected EOF", 0);
		end
		Text::i := 0;
		return STRING;
	end
	i := findop(c);
	if (i) return i;
	bc::0 := c/16 + (c/16 > 9 -> 'A'-10 : '0');
	bc::1 := c mod 16 + (c mod 16 > 9 -> 'A'-10 : '0');
	bc::2 := 0;
	fatal(packed"bad input character", bc);
end


synch(tok) do
	ie (tok) do
		while (Token \= tok) do
			Token := scan();
			if (Token = ENDFILE)
				fatal(packed"EOF found in error recovery", 0);
		end
	end
	else do
		Token := scan();
	end
end


findivar(tbl, base, lim, name, rpt) do var i, s;
	i := base + SYMREC;
	while (i < lim /\ tbl[i+SFLAGS] & IVAR) do
		s := tbl[i+SNAME];
		if (s::0 = name::0 /\ strequ(name, s)) return i;
		i := i+SYMREC;
	end
	if (rpt) error(packed"undefined symbol", name);
	return %1;
end


extend(h, l) do var v;
	v := h<<8 | l;
	if (v > 32767) v := v-K64;
	return v;
end


extendcache(cache, len) do var i, y, r;
	for (i=len-1, -1, -1) do
		r := @cache::(i*SR2);
		y := @cache[i*SYMREC];
		! Reverse order is important. Otherwise, source
		! fields may be overwritten!
		y[SFLAGS] := extend(r::SFH, r::SFL);
		y[SSIZE] := extend(r::SZH, r::SZL);
		y[SVAL] := extend(r::SVH, r::SVL);
		y[SNAME] := 0;
	end
end


rdclsdir(dirname, reqd) do
	var	i, j, k, lib;
	var	magic::4;
	var	rf;
	var	n, nbase;
	var	nlt, nlen;
	var	nb::2;

	rf := packed"corrupt CLASSDIR";
	lib := t.open(dirname, T3X.OREAD);
	if (lib = %1) do
		if (\reqd) return 0;
		fatal(packed"class not available", dirname);
	end
	if (	t.read(lib, magic, 4) \= 4 \/
		t.memcomp(magic, packed"CDIR", 4)
	)
		fatal(rf, packed"#1");
	if (t.read(lib, nb, 2) \= 2) fatal(rf, packed"#2");
	n := nb::1 << 8 | nb::0;
	if (Lt + n*SYMREC >= CACHESPACE)
		fatal(packed"class directory overflow", 0);
	nlen := n*SYMREC;
	nlt := Lt + nlen;
	if (t.read(lib, @Libcache[Lt], nlen*2) \= nlen*2)
		fatal(rf, packed"#3");
	if (\TUNE16) extendcache(@Libcache[Lt], n);
	nbase := Nt;
	Nt := Nt + t.read(lib, @Names::Nt, NSPACE-Nt);
	if (Nt >= NSPACE \/ Nt < nbase)
		fatal(packed"name space overflow (rdclsdir)", 0);
	j := Lt;
	k := nbase;
	for (i=nbase, Nt) do
		if (Names::i = 0) do
			Libcache[j+SNAME] := @Names::k;
			k := i+1;
			j := j+SYMREC;
		end
		if (j > nlt) leave;
	end
	if (j \= nlt) fatal(rf, packed"#4");
	Lt := nlt;
	t.close(lib);
end


wrclsdir() do
	const	WBUFLEN = 256;
	var	k, i, j, f, lib;
	var	buf::WBUFLEN, wp;
	var	wf, pub;
	var	nb::2, p, q;

	wf := packed"error writing CLASSDIR";
	k := 0;
	pub := 0;
	for (i=0, Lt, SYMREC) do
		f := Libcache[i+SFLAGS];
		if (f & CLSS) pub := f & PUBLC;
		if (\(f & SYS))
			Libcache[i+SFLAGS] := f & ~LINKD | (f->EXTRN: 0);
		if (pub) k := k+1;
	end
	if (k = 0) return 0;
	lib := t.open(packed"CLASSDIR", T3X.OWRITE);
	if (lib = %1) fatal(packed"could not update CLASSDIR", 0);
	nb::0 := k & 255;
	nb::1 := k >> 8;
	if (	t.write(lib, packed"CDIR", 4) \= 4 \/
		t.write(lib, nb, 2) \= 2
	)
		fatal(wf, packed"#1");
	pub := 0;
	wp := 0;
	for (i=0, Lt, SYMREC) do
		if (Libcache[i+SFLAGS] & CLSS)
			pub := Libcache[i+SFLAGS] & PUBLC;
		if (pub) do
			if (wp + SR2 >= WBUFLEN) do
				if (t.write(lib, buf, wp) \= wp)
					fatal(wf, packed"#2");
				wp := 0;
			end
			ie (TUNE16) do
				t.memcopy(@buf::wp, @Libcache[i], SR2);
			end
			else do
				p := @buf::wp;
				q := @Libcache[i];
				p::SNL := 0;
				p::SNH := 0;
				p::SVL := q[SVAL] & 255;
				p::SVH := q[SVAL] >> 8;
				p::SZL := q[SSIZE] & 255;
				p::SZH := q[SSIZE] >> 8;
				p::SFL := q[SFLAGS] & 255;
				p::SFH := q[SFLAGS] >> 8;
			end
			wp := wp+SR2;
		end
	end
	pub := 0;
	for (i=0, Lt, SYMREC) do
		if (Libcache[i+SFLAGS] & CLSS)
			pub := Libcache[i+SFLAGS] & PUBLC;
		if (\pub) loop;
		j := length(Libcache[i+SNAME]);
		if (j + wp >= WBUFLEN) do
			if (t.write(lib, buf, wp) \= wp)
				fatal(wf, packed"#3");
			wp := 0;
		end
		t.memcopy(@buf::wp, Libcache[i+SNAME], j+1);
		wp := wp+j+1;
	end
	if (wp /\ t.write(lib, buf, wp) \= wp)
		fatal(wf, packed"#4");
	t.close(lib);
end


findintclass(name) do var i, j, s;
	for (i=0, Lt, SYMREC) do
		s := Libcache[i+SNAME];
		if (	Libcache[i+SFLAGS] & CLSS /\
			s::0 = name::0 /\ strequ(name, s)
		)
			return i;
	end
	return %1;
end


mkclasspath(dest, base, dir) do var i, j, sfx;
	sfx := packed".dir";
	j := 0;
	i := 0; while (base::i) do
		dest::j := base::i;
		i := i+1; j := j+1;
	end
	dest::j := '/'; j := j+1;
	i := 0; while (dir::i) do
		dest::j := dir::i;
		i := i+1; j := j+1;
	end
	i := 0; while (sfx::i) do
		dest::j := sfx::i;
		i := i+1; j := j+1;
	end
	dest::j := 0;
end


findclass(name, reqd) do
	var	cl, i;
	var	cldir::TEXTLEN+1;

	cl := findintclass(name);
	if (cl \= %1 \/ \reqd) return cl;
	i := 0;
	while (cl = %1) do
		if (\Classpaths[i]) leave;
		if (Classpaths[i]::0) do
			mkclasspath(cldir, Classpaths[i], name);
			rdclsdir(cldir, 0);
			cl := findintclass(name);
		end
		i := i+1;
	end
	if (cl = %1) fatal(packed"class not available", name);
	return cl;
end


findredclass(ctx, name) do var i, cl;
	i := ctx-SYMREC;
	while (i >= 0) do
		if (\(Symbols[i+SFLAGS] & REDIR)) leave;
		cl := Symbols[i+SVAL];
		if (strequ(Libcache[cl+SNAME], name)) return cl;
		i := i - SYMREC;
	end
	return %1;
end


findclassref(name, rpt) do var k;
	k := Classctx = %1-> %1: findredclass(Classctx, name);
	if (k = %1 /\ Modctx \= %1) k := findredclass(Modctx, name);
	if (rpt /\ k = %1) error(packed"class not available", name);
	return k;
end


findsym(name, rpt) do var i, s;
	i := St-SYMREC;
	while (i >= 0) do
		s := Symbols[i+SNAME];
		if (s::0 = name::0 /\ strequ(name, s))
			return i;
		i := i-SYMREC;
	end
	if (rpt) error(packed"undefined symbol", name);
	return %1;
end


rptdecls(from, to) do var i, s;
	i := from;
	while (i < to) do
		s := @Symbols[i];
		if (s[SFLAGS] & PROTO)
			error(packed"undefined procedure", s[SNAME]);
		i := i+SYMREC;
	end
end


newsym(name, size, val, flags) do var k, sp;
	k := length(name);
	if (findclass(name, 0) \= %1)
		error(packed"duplicate symbol (class conflict)", name);
	sp := findsym(name, 0);
	if (name::0 \= '*' /\ name::0 \= '+' /\ sp \= %1) do
		if (	flags & (PROC|PROTO) \= PROC \/ 
			Symbols[sp+SFLAGS] & (PROC|PROTO) \= (PROC|PROTO)
		) do
			error(packed"duplicate symbol", name);
			return sp;
		end
	end
	if (St >= SYMBSPACE) fatal(packed"too many symbols", name);
	if (Nt + k >= NSPACE) fatal(packed"out of name space", name);
	strcpy(@Names::Nt, name);
	Symbols[St+SNAME] := @Names::Nt;
	Nt := Nt+k+1;
	Symbols[St+SSIZE] := size;
	Symbols[St+SVAL] := val;
	Symbols[St+SFLAGS] := flags |
		((Classctx \= %1)-> IVAR: 0);
	St := St+SYMREC;
	return St-SYMREC;
end


newlab() do
	if (\Label) fatal(packed"out of labels", 0);
	Label := Label+1;
	return Label-1;
end


flush() do
	t.write(T3X.SYSOUT, Obuf, Obp);
	Obp := 0;
end


commit() do var i;
	for (i=0, LL) do
		if (Obp >= OBUFL) flush();
		Obuf::Obp := Last[i];
		Obp := Obp + 1;
	end
end


gen0(instr) do
	commit();
	Last[0] := instr;
	LL := 1;
end


gen1(instr, arg) do
	commit();
	Last[0] := instr;
	Last[1] := arg & 255;
	Last[2] := arg>>8;
	LL := 3;
end


gen2(instr, a1, a2) do
	commit();
	Last[0] := instr;
	Last[1] := a1 & 255;
	Last[2] := a1>>8;
	Last[3] := a2 & 255;
	Last[4] := a2>>8;
	LL := 5;
end


stack(n) if (n) do
		gen1(tcode.ISTACK, n);
		if (Debug) gen0(tcode.ICKSTK);
	end


mtdadj(n) return n<0-> n: n+Method_adjust;


linkage(type, base, k) do
	var i, j, lab, s::TEXTLEN*2+1;

	ie (type = tcode.IPUB) do
		i := strcpy(s, Symbols[Classctx+SNAME]);
		s::i := '_'; i := i+1;
		strcpy(@s::i, Symbols[k+SNAME]);
		lab := Symbols[k+SVAL];
	end
	else do
		i := strcpy(s, Libcache[base+SNAME]);
		s::i := '_'; i := i+1;
		strcpy(@s::i, Libcache[k+SNAME]);
		if (\Xlabel) fatal(packed"too many external labels", 0);
		lab := Xlabel;
		Xlabel := Xlabel+1;
		Libcache[k+SVAL] := lab;
	end
	j := length(s);
	gen2(type, lab, j);
	for (i=0, j) gen0(s::i);
end


loadsym(k) do var f, z;
	f := Symbols[k+SFLAGS];
	if (f & PROC) return 0;
	z := (f & OBJCT)-> Libcache[Symbols[k+SSIZE]+SSIZE]:
		Symbols[k+SSIZE];
	ie (k = Selfobj) do
		gen0(tcode.ISELF);
	end
	else ie (f & GLOBL) do
		ie (f & IVAR)
			gen1(z-> tcode.ILDIV: tcode.ILDI, Symbols[k+SVAL]);
		else
			gen1(z-> tcode.ILDGV: tcode.ILDG, Symbols[k+SVAL]);
	end
	else do
		gen1(z-> tcode.ILDLV: tcode.ILDL, mtdadj(Symbols[k+SVAL]));
	end
end


savesym(k) do var f;
	f := Symbols[k+SFLAGS];
	ie (f & GLOBL)
		gen1(f & IVAR-> tcode.ISAVI: tcode.ISAVG, Symbols[k+SVAL]);
	else
		gen1(tcode.ISAVL, mtdadj(Symbols[k+SVAL]));
end


tcond(n, flags) do
	ie (flags) do
		gen1(flags=1-> tcode.INBRT: tcode.INBRF, n);
		gen0(tcode.IPOP);
	end
	else do
		gen1(tcode.IBRF, n);
	end
end


fcond(y, n) do
	gen1(tcode.IJUMP, y);
	gen1(tcode.ICLAB, n);
end


econd(y) gen1(tcode.ICLAB, y);


match(tok, m) do
	ie (tok = Token)
		Token := scan();
	else
		error(m, 0);
end


xsemi() match(SEMI, packed"missing ';'");

xlparen() match(LPAREN, packed"missing '('");

xrparen() match(RPAREN, packed"missing ')'");

xrbrack() match(RBRACK, packed"missing ']'");

xcomma(swallow) ie (Token = COMMA)
		if (swallow) Token := scan();
	else
		error(packed"missing ','", 0);

equsign() ie (Token \= BINOP \/ Op \= Sym_equ)
		error(packed"missing '='", 0);
	else
		Token := scan();


forcesym(msg) ie (Token = SYMBOL) do
		return 0;
	end
	else do
		error(msg-> msg: packed"symbol expected", 0);
		synch(SEMI); Token := scan();
		return %1;
	end


classconst(comp) do var k, v;
	k := findclassref(Text, 0);
	if (k = %1) error(packed"undefined symbol", Text);
	Token := scan();
	if (k = %1) return %1;
	ie (Token = DOT) do
		Token := scan();
		if (forcesym(0)) return %1;
		k := findivar(Libcache, k, Lt, Text, 0);
		if (\(Libcache[k+SFLAGS] & CNST))
			error(packed"not a constant", Text);
		ie (k \= %1) do
			v := Libcache[k+SVAL];
			if (comp) gen1(tcode.INUM, v);
		end
		else do
			error(packed"no such public constant", Text);
		end
		Token := scan();
	end
	else do
		v := Libcache[k+SSIZE];
		if (comp) gen1(tcode.INUM, v);
	end
	return v;
end


constfac() do var k, v;
	v := 1;
	ie (Token = BINOP /\ Op = Sym_sub) do
		Token := scan();
		return -constfac();
	end
	else ie (Token = UNOP /\ Op = Sym_not) do
		Token := scan();
		return ~constfac();
	end
	else ie (Token = NUMBER) do
		v := Value;
		Token := scan();
	end
	else ie (Token = SYMBOL) do
		k := findsym(Text, 0);
		ie (k \= %1) do
			ie (Symbols[k+SFLAGS] & CNST)
				v := Symbols[k+SVAL];
			else ie (Symbols[k+SFLAGS] & CLSS) do
				v := Symbols[k+SSIZE];
			end
			else do
				error(packed"not a constant", Text);
			end
			Token := scan();
		end
		else do
			v := classconst(0);
		end
	end
	else do
		error(packed"constant value expected", 0);
	end
	return v;
end


constval() do var v, o;
	v := constfac();
	while (Token = BINOP) do
		o := Op;
		Token := scan();
		ie (Op = Sym_add)
			v := v + constfac();
		else ie (Op = Sym_mul)
			v := v * constfac();
		else ie (Op = Sym_bar)
			v := v | constfac();
		else
			leave;
	end
	return v;
end


pcall(k, ind) do
	var	n, d[5], dl, i;

	n := 0;
	ie (k = %1) do
		error(packed"call of non-procedure", 0);
	end
	else do
		ie (ind) do
			ie (Symbols[k+SFLAGS] & PROC /\
				\(Symbols[k+SFLAGS] & PUBLC)
			)
				ind := 0;
			else if ((Symbols[k+SFLAGS] &
				 (CNST|CLSS|OBJCT|PUBLC)) \/
				Symbols[k+SSIZE]
			)
				error(packed"bad indirect call",
					Symbols[k+SNAME]);
		end
		else do
			if (	\(Symbols[k+SFLAGS] & PROC) \/
				Symbols[k+SFLAGS] & PUBLC
			)
				error(packed"call of non-procedure",
					Symbols[k+SNAME]);
		end
	end
	xlparen();
	if (ind) do
		for (i=0, 5) d[i] := Last[i];
		dl := LL; LL := 0;
	end
	if (Token \= RPAREN) while (1) do
		expr();
		n := n+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	if (\ind /\ k >= 0 /\ Symbols[k+SSIZE] \= n)
		error(packed"wrong number of arguments", Symbols[k+SNAME]);
	if (k >= 0) do
		ie (ind) do
			commit();
			for (i=0, 5) Last[i] := d[i];
			LL := dl;
			gen0(tcode.ICALR);
		end
		else ie (Symbols[k+SFLAGS] & SYS) do
			gen1(tcode.ISYS, Symbols[k+SVAL]);
		end
		else ie (Symbols[k+SFLAGS] & EXTRN) do
			gen1(tcode.ICALX, Symbols[k+SVAL]);
		end
		else do
			gen1(tcode.ICALL, Symbols[k+SVAL]);
		end
		gen1(tcode.ICLEAN, n);
	end
	xrparen();
end


strlit(s, ext) do var i, l, k;
	if (Packstr) ext := 0;
	k := length(s);
	l := newlab();
	gen1(tcode.IDLAB, l);
	gen1(ext-> tcode.ISTR: tcode.IPSTR, k);
	for (i=0, k) gen0(s::i);
	Token := scan();
	return l;
end


bytevec() do var tbl[MAXTBL], v, i, p;
	Token := scan();
	if (Token = STRING) return strlit(Text, 0);
	if (Token \= LBRACK) do
		error(packed"packed string or vector expected", 0);
		synch(0);
		return 0;
	end
	p := 0;
	Token := scan();
	while (Token = SYMBOL \/ Token = NUMBER) do
		tbl[p] := constval();
		if (tbl[p] > 255 \/ tbl[p] < -128)
			error(packed"vector member too big", 0);
		if (p >= MAXTBL) fatal(packed"table too big", 0);
		p := p+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	if (\p) error(packed"empty table", 0);
	v := newlab();
	gen1(tcode.IDLAB, v);
	gen1(tcode.IPSTR, p);
	for (i=0, p) gen0(tbl[i]);
	xrbrack();
	return v;
end


table() do
	var	tv[MAXTBL], tf[MAXTBL], p, v, i, k;

	p := 0;
	Token := scan();
	while (	Token = SYMBOL \/ Token = NUMBER \/
		Token = LBRACK \/ Token = STRING \/
		Token = ADDROP \/ Token = KPACKED \/
		Token = LPAREN \/
		Token = BINOP /\ Op = Sym_sub \/
		Token = UNOP /\ Op = Sym_not
	) do
		ie (	Token = SYMBOL \/ Token = NUMBER \/
			Token = BINOP /\ Op = Sym_sub \/
			Token = UNOP /\ Op = Sym_not
		) do
			tv[p] := constval();
			tf[p] := tcode.IDATA;
		end
		else ie (Token = STRING) do
			tv[p] := strlit(Text, 1);
			tf[p] := tcode.IDREF;
		end
		else ie (Token = KPACKED) do
			tv[p] := bytevec();
			tf[p] := tcode.IDREF;
		end
		else ie (Token = ADDROP) do
			Token := scan();
			ie (Token = SYMBOL) do
				k := findsym(Text, 1);
			end
			else do
				error(packed"symbol expected", 0);
				k := %1;
			end
			if (k >= 0) do
				ie (Symbols[k+SFLAGS] & (CNST|SYS)) error(
					packed"cannot take @const or @syscall",
					Text);
				else ie (Symbols[k+SFLAGS] & PROC)
					tf[p] := tcode.ICREF;
				else ie (Symbols[k+SFLAGS] & GLOBL = GLOBL)
					tf[p] := tcode.IDREF;
				else
					error(packed"unknown address", Text);
			end
			tv[p] := Symbols[k+SVAL];
			Token := scan();
		end
		else ie (Token = LPAREN) do
			tv[p] := newlab();
			tf[p] := tcode.IDLAB;
			expr();
			gen1(tcode.ISAVG, tv[p]);
		end
		else do
			tv[p] := table();
			tf[p] := tcode.IDREF;
		end
		if (p >= MAXTBL) fatal(packed"table too big", 0);
		p := p+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	v := newlab();
	gen1(tcode.IDLAB, v);
	if (\p) error(packed"empty table", 0);
	for (i=0, p) do
		ie (tf[i] = tcode.IDLAB) do
			gen1(tcode.IDLAB, tv[i]);
			gen1(tcode.IDATA, 0);
		end
		else do
			gen1(tf[i], tv[i]);
		end
	end
	xrbrack();
	return v;
end


srchmsg(base, name) do var m;
	m := findivar(Libcache, base, Lt, name, 0);
	if (m \= %1 /\ Libcache[m+SFLAGS] & (EXTRN|LINKD) = EXTRN) do
		linkage(tcode.IEXT, base, m);
		Libcache[m+SFLAGS] := Libcache[m+SFLAGS] | LINKD;
	end
	return m;
end


findmsg(k, ptbl) do var m, tbl;
	m := %1;
	ie (Symbols[k+SSIZE] = Classctx) do
		tbl := Symbols;
		m := findivar(Symbols, Classctx, St, Text, 0);
	end
	else do
		tbl := Libcache;
		m := srchmsg(Symbols[k+SSIZE], Text);
	end
	if (m \= %1 /\ \(tbl[m+SFLAGS] & PROC)) m := %1;
	ptbl[0] := tbl;
	return m;
end


sendmsg(k, cl) do
	var	m, n;
	var	d[5], dl, i;
	var	tbl;

	for (i=0, 5) d[i] := Last[i];
	dl := LL; LL := 0;
	Token := scan();
	if (k = %1 \/ Symbols[k+SFLAGS] & OBJCT \= OBJCT)
		error(packed"invalid object", k=%1-> 0: Symbols[k+SNAME]);
	if (forcesym(packed"method name expected")) return 0;
	ie (cl \= %1) do
		m := srchmsg(cl, Text);
		tbl := Libcache;
	end
	else do
		m := k=%1-> %1: findmsg(k, @tbl);
	end
	if (m = %1 /\ k \= %1) do
		error(packed"cannot answer message", Text);
	end
	Token := scan();
	if (Token \= LPAREN) do
		synch(SEMI); return 0;
	end
	xlparen();
	n := 0;
	if (Token \= RPAREN) while (1) do
		expr();
		n := n+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	if (m \= %1 /\ tbl[m+SSIZE] \= n)
		error(packed"wrong number of arguments", tbl[m+SNAME]);
	commit();
	for (i=0, 5) Last[i] := d[i];
	LL := dl;
	if (m >= 0) do
		gen1(tbl[m+SFLAGS] & SYS-> tcode.ISYS:
			tbl[m+SFLAGS] & EXTRN-> tcode.ICALX:
			tcode.ICALL, tbl[m+SVAL]);
		gen1(tcode.ICLEAN, n+1);
	end
	xrparen();
end


address() do var tok, k;
	if (Token \= SYMBOL) do
		error(packed"symbol expected", 0);
		return %1;
	end
	k := findsym(Text, 0);
	ie (k = %1) do
		classconst(1);
		return %1;
	end
	else do
		Token := scan();
		ie (Symbols[k+SFLAGS] & CNST)
			gen1(tcode.INUM, Symbols[k+SVAL]);
		else ie (Symbols[k+SFLAGS] & CLSS)
			gen1(tcode.INUM, Symbols[k+SSIZE]);
		else
			loadsym(k);
	end
	while (Token = LBRACK \/ Token = BYTEOP) do
		if (k >= 0 /\ (Symbols[k+SFLAGS] & (CNST|PROC|OBJCT|CLSS)))
			error(packed"bad subscript", Symbols[k+SNAME]);
		tok := Token;
		Token := scan();
		ie (tok = BYTEOP) do
			factor();
			gen0(tcode.IDREFB);
			gen0(tcode.IINDB);
			return %1;
		end
		else do
			expr();
			gen0(tcode.IDEREF);
			gen0(tcode.IIND);
			xrbrack();
		end
		k := %1;
	end
	return k;
end


sendop() do var p, cl;
	Token := scan();
	xlparen();
	if (forcesym(packed"missing variable in SEND")) return %1;
	p := findsym(Text, 1);
	if (p < 0) do
		synch(SEMI); return 0;
	end
	if (Symbols[p+SFLAGS] & (CNST|CLSS|OBJCT|PROC)) do
		error(packed"bad pointer in SEND", Symbols[p+SNAME]);
		synch(SEMI); return 0;
	end
	Token := scan();
	xcomma(1);
	if (forcesym(packed"missing class in SEND")) return %1;
	cl := findclassref(Text, 1);
	if (cl < 0) do
		synch(SEMI); return 0;
	end
	Token := scan();
	xcomma(0);
	loadsym(p);
	sendmsg(0, cl);
	xrparen();
end


factor() do var o, k, ind;
	ie (Token = NUMBER) do
		gen1(tcode.INUM, Value);
		Token := scan();
	end
	else ie (Token = SYMBOL \/ Token = KCALL) do
		ind := 0;
		if (Token = KCALL) do
			Token := scan(); ind := 1;
		end
		k := address();
		ie (Token = DOT) do
			sendmsg(k, %1);
		end
		else ie (Token = LPAREN)
			pcall(k, ind);
		else if (k >= 0 /\ Symbols[k+SFLAGS] & PROC \/ ind)
			error(packed"incomplete procedure call",
				Symbols[k+SNAME]);
	end
	else ie (Token = KSEND) do
		sendop();
	end
	else ie (Token = STRING) do
		gen1(tcode.ILDLAB, strlit(Text, 1));
	end
	else ie (Token = KPACKED) do
		gen1(tcode.ILDLAB, bytevec());
	end
	else ie (Token = LBRACK) do
		gen1(tcode.ILDLAB, table());
	end
	else ie (Token = ADDROP) do
		Token := scan();
		k := address();
		ie (k >= 0) do
			if (\(Symbols[k+SFLAGS] & PROC)) LL := 0;
			ie (Symbols[k+SFLAGS] & (CNST|SYS))
				error(packed"cannot take @const or @syscall",
					Symbols[k+SNAME]);
			else ie (Symbols[k+SFLAGS] & PROC)
				gen1(tcode.ILDLAB, Symbols[k+SVAL]);
			else ie (Symbols[k+SFLAGS] & GLOBL)
				gen1(Symbols[k+SFLAGS] & IVAR-> tcode.ILDIV:
					tcode.ILDGV, Symbols[k+SVAL]);
			else
				gen1(tcode.ILDLV, mtdadj(Symbols[k+SVAL]));
		end
		else do
			ie (Last[0] = tcode.IIND \/ Last[0] = tcode.IINDB)
				LL := 0;
			else
				error(packed"bad lvalue", 0);
		end
	end
	else ie (Token = BINOP) do
		if (Op \= Sym_sub) error(packed"bad unary operator", 0);
		Token := scan();
		factor();
		gen0(tcode.INEG);
	end
	else ie (Token = UNOP) do
		o := Op;
		Token := scan();
		factor();
		gen0(Ops[o][OCODE]);
	end
	else ie (Token = LPAREN) do
		Token := scan();
		expr();
		xrparen();
	end
	else do
		error(packed"bad expression", 0);
		synch(0);
	end
end


emitop(stk, sp) do
	sp := sp-1;
	gen0(Ops[stk[sp]][OCODE]);
	return sp;
end


binary() do var stk[5], sp;
	sp := 0;
	factor();
	while (Token = BINOP) do
		while (sp /\ Ops[Op][OPREC] <= Ops[stk[sp-1]][OPREC])
			sp := emitop(stk, sp);
		if (sp >= 5) fatal(packed"binary(): stack overflow", 0);
		stk[sp] := Op; sp := sp+1;
		Token := scan();
		factor();
	end
	while (sp) sp := emitop(stk, sp);
end


conj() do var e;
	e := 0;
	binary();
	if (Token = CONOP) e := newlab();
	while (Token = CONOP) do
		Token := scan();
		tcond(e, 2);
		binary();
	end
	if (e) econd(e);
end


disj() do var e;
	e := 0;
	conj();
	if (Token = DISOP) e := newlab();
	while (Token = DISOP) do
		Token := scan();
		tcond(e, 1);
		conj();
	end
	if (e) econd(e);
end


expr() do var y, n;
	disj();
	if (Token = COND) do
		y := newlab();
		n := newlab();
		Token := scan();
		tcond(n, 0);
		expr();
		fcond(y, n);
		match(COLON, packed"missing ':'");
		expr();
		econd(y);
	end
end


ie_stmt(alt) do var y, n;
	n := newlab();
	if (alt) y := newlab();
	Token := scan();
	xlparen();
	expr();
	xrparen();
	tcond(n, 0);
	stmt();
	if (alt) do
		fcond(y, n);
		match(KELSE, packed"missing 'ELSE'");
		stmt();
	end
	econd(alt-> y: n);
end


while_stmt() do var s, e, levl;
	s := Startlab;
	e := Endlab;
	levl := Looplevl;
	Startlab := newlab();
	Endlab := newlab();
	Looplevl := Locladdr;
	Token := scan();
	xlparen();
	gen1(tcode.ICLAB, Startlab);
	expr();
	tcond(Endlab, 0);
	xrparen();
	stmt();
	gen1(tcode.IJUMP, Startlab);
	econd(Endlab);
	Startlab := s;
	Endlab := e;
	Looplevl := levl;
end


for_stmt() do var k, step, r, s, e, levl;
	step := 1;
	s := Startlab;
	e := Endlab;
	levl := Looplevl;
	Startlab := newlab();
	r := newlab();
	Endlab := newlab();
	Looplevl := Locladdr;
	Token := scan();
	xlparen();
	ie (Token = SYMBOL) do
		k := findsym(Text, 1);
		if (k \= %1 /\ Symbols[k+SFLAGS] & (CNST|PROC|CLSS|OBJCT))
			error(packed"bad lvalue", Text);
		if (k \= %1 /\ Symbols[k+SSIZE])
			error(packed"index may not be a vector", Text);
	end
	else do
		k := %1;
	end
	match(SYMBOL, packed"symbol name expected");
	equsign();
	expr();
	if (k \= %1) savesym(k);
	xcomma(1);
	gen1(tcode.ICLAB, r);
	if (k \= %1) loadsym(k);
	expr();
	if (Token = COMMA) do
		Token := scan();
		step := constval();
	end
	xrparen();
	gen1(step>0-> tcode.IUNEXT: tcode.IDNEXT, Endlab);
	stmt();
	gen1(tcode.ICLAB, Startlab);
	ie (Symbols[k+SFLAGS] & GLOBL)
		ie (Symbols[k+SFLAGS] & IVAR)
			gen2(tcode.IINCI, Symbols[k+SVAL], step);
		else
			gen2(tcode.IINCG, Symbols[k+SVAL], step);
	else
		gen2(tcode.IINCL, mtdadj(Symbols[k+SVAL]), step);
	gen1(tcode.IJUMP, r);
	gen1(tcode.ICLAB, Endlab);
	Startlab := s;
	Endlab := e;
	Looplevl := levl;
end


asg_or_call() do var k;
	if (forcesym(packed"bad statement")) return 0;
	k := address();
	ie (Token = ASSIGN) do
		Token := scan();
		ie (k \= %1) do
			ie (Symbols[k+SFLAGS] & (CNST|PROC|OBJCT|CLSS))
				error(packed"bad lvalue", Symbols[k+SNAME]);
			else if (Symbols[k+SSIZE])
				error(packed"missing subscript",
					Symbols[k+SNAME]);
			LL := 0;
			expr();
			savesym(k);
		end
		else do
			k := Last[0]; LL := 0;
			if (k \= tcode.IIND /\ k \= tcode.IINDB)
				error(packed"bad lvalue", 0);
			expr();
			gen0(k=tcode.IIND-> tcode.ISTORE: tcode.ISTORB);
		end
		xsemi();
	end
	else ie (Token = LPAREN) do
		pcall(k, 0);
		gen0(tcode.IPOP);
		xsemi();
	end
	else ie (Token = DOT) do
		sendmsg(k, %1);
		gen0(tcode.IPOP);
		xsemi();
	end
	else do
		error(packed"bad statement", 0);
		synch(SEMI);
		xsemi();
	end
end


defr5proc(name, arity) do var i, j;
	newsym(name, arity, Xlabel, PROC|EXTRN);
	if (\Xlabel) fatal(packed"too many external labels", 0);
	Xlabel := Xlabel+1;
	j := length(name);
	gen2(tcode.IEXT, Xlabel-1, j);
	for (i=0, j) gen0(name::i);
end


metacmd() do var badsyn;
	badsyn := 0;
	Token := scan();
	ie (Token \= SYMBOL) do
		badsyn := 1;
	end
	else ie (strequ(Text, packed"classpath")) do
		Token := scan();
		ie (Token \= STRING) do
			badsyn := 1;
		end
		else do
			strcpy(Usrcp, Text);
			Classpaths[4] := Usrcp;
			Token := scan();
		end
	end
	else ie (Text::0 = 'l' /\ \Text::2) do
		Token := scan();
		ie (Token \= NUMBER) badsyn := 1;
		else Token := scan();
		Line := Value;
		ie (Token \= STRING) do
			badsyn := 1;
		end
		else do
			if (\badsyn) strcpy(File, Text);
			Token := scan();
		end
	end
	else ie (strequ(Text, packed"debug")) do
		Token := scan();
		Debug := 1;
	end
	else ie (strequ(Text, packed"packstrings")) do
		Token := scan();
		Packstr := 1;
	end
	else ie (strequ(Text, packed"r5")) do
		Token := scan();
		defr5proc(packed"open", 2);
		defr5proc(packed"close", 1);
		defr5proc(packed"erase", 1);
		defr5proc(packed"select", 2);
		defr5proc(packed"reads", 2);
		defr5proc(packed"writes", 1);
		defr5proc(packed"newline", 0);
		defr5proc(packed"aton", 1);
		defr5proc(packed"ntoa", 2);
		defr5proc(packed"pack", 2);
		defr5proc(packed"unpack", 2);
		defr5proc(packed"readpacked", 3);
		defr5proc(packed"writepacked", 3);
		defr5proc(packed"reposition", 4);
		defr5proc(packed"rename", 2);
		defr5proc(packed"memcopy", 3);
		defr5proc(packed"memcomp", 3);
	end
	else do
		Token := scan();
		badsyn := 1;
	end
	if (badsyn) error(packed"invalid # syntax", 0);
	xsemi();
end


stmt() do
	if (Debug) gen1(tcode.ILINE, Line);
	ie (Token = KDO) do
		return compound();
	end
	else ie (Token = KIE) do
		ie_stmt(1);
	end
	else ie (Token = KIF) do
		ie_stmt(0);
	end
	else ie (Token = KWHILE) do
		while_stmt();
	end
	else ie (Token = KFOR) do
		for_stmt();
	end
	else ie (Token = KLEAVE) do
		Token := scan();
		xsemi();
		if (\Endlab) error(packed"'LEAVE' in wrong context", 0);
		stack(-(Locladdr - Looplevl));
		gen1(tcode.IJUMP, Endlab);
	end
	else ie (Token = KLOOP) do
		Token := scan();
		xsemi();
		if (\Startlab) error(packed"'LOOP' in wrong context", 0);
		stack(-(Locladdr - Looplevl));
		gen1(tcode.IJUMP, Startlab);
	end
	else ie (Token = KRETURN) do
		Token := scan();
		ie (Token = SEMI)
			gen1(tcode.INUM, 0);
		else
			expr();
		xsemi();
		if (\Exitlab) error(packed"'RETURN' in wrong context", 0);
		gen0(tcode.IPOP);
		stack(-Locladdr+1);
		gen1(tcode.IJUMP, Exitlab);
		return 1;
	end
	else ie (Token = KHALT) do
		Token := scan();
		ie (Token = SEMI)
			gen1(tcode.IHALT, 0);
		else
			gen1(tcode.IHALT, constval());
		xsemi();
	end
	else ie (Token = SEMI) do
		Token := scan();
	end
	else ie (Token = KELSE) do
		Token := scan();
		error(packed"'ELSE' with no matching 'IE'", 0);
	end
	else ie (Token = KCALL) do
		factor();
		gen0(tcode.IPOP);
		xsemi();
	end
	else ie (Token = KSEND) do
		sendop();
		gen0(tcode.IPOP);
		xsemi();
	end
	else ie (Token = METAOP) do
		metacmd();
	end
	else do
		asg_or_call();
	end
	return 0;
end


cleanup() do var k, la;
	la := Locladdr;
	k := findsym(packed"*", 0);
	if (k \= %1) do
		St := k;
		Nt := Symbols[k+SSIZE];
		Locladdr := Symbols[k+SVAL];
	end
	stack(-(la - Locladdr));
end


compound() do var ot, rv;
	ot := Locladdr;
	newsym(packed"*", Nt, Locladdr, 0);
	Token := scan();
	while (	Token = KVAR \/ Token = KCONST \/ Token = KSTRUCT \/
		Token = METAOP \/ Token = KOBJECT
	)
		declaration(1);
	stack(Locladdr - ot);
	rv := 0;
	while (Token \= KEND) do
		if (eof()) fatal(packed"unexpected EOF", 0);
		rv := stmt();
	end
	Token := scan();
	cleanup();
	return rv;
end


adjust(k, n) do
	n := n+3;
	while (k < St) do
		Symbols[k+SVAL] := Symbols[k+SVAL]-n;
		k := k+SYMREC;
	end
end


dumpsym(k, local) do var i, j, s;
	s := Symbols[k+SNAME];
	j := length(s);
	ie (local)
		gen1(tcode.ILSYM, mtdadj(Symbols[k+SVAL]));
	else ie (Symbols[k+SFLAGS] & IVAR)
		gen1(tcode.IISYM, Symbols[k+SVAL]);
	else
		gen1(tcode.IGSYM, Symbols[k+SVAL]);
	gen0(j);
	gen0(j>>8);
	for (i=0, j) gen0(s::i);
end


procdef(pub) do
	var	p, otbl, i, k, pb, lv, n, addr, rv;

	n := 0;
	addr := pub->1:2;
	Method_adjust := pub->1:0;
	p := findsym(Text, 0);
	Exitlab := newlab();
	k := newsym(Text, 0, 0, PROC | LINKD | (pub->PUBLC:0));
	lv := newsym(packed"*", Nt, 1, 0)+SYMREC;
	Token := scan();
	xlparen();
	pb := St;
	if (Token \= RPAREN) while (1) do
		if (Token = SYMBOL) do
			newsym(Text, 0, addr, 0);
			addr := addr+1;
		end
		match(SYMBOL, packed"argument name expected");
		n := n+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	Symbols[k+SSIZE] := n;
	ie (p \= %1 /\ (Symbols[p+SFLAGS] & PROTO)) do
		if (Symbols[p+SSIZE] \= n)
			error(packed"argument count mismatch",
				Symbols[k+SNAME]);
		Symbols[k+SVAL] := Symbols[p+SVAL];
		Symbols[p+SFLAGS] := Symbols[p+SFLAGS] & IVAR;
		Symbols[p+SNAME] := packed"?";
	end
	else do
		Symbols[k+SVAL] := newlab();
	end
	if (pub /\ Classctx \= %1 /\ Symbols[Classctx+SFLAGS] & PUBLC)
		linkage(tcode.IPUB, 0, k);
	gen1(tcode.ICLAB, Symbols[k+SVAL]);
	gen0(tcode.IHDR);
	if (pub) gen0(tcode.ICCTX);
	if (Debug) gen0(tcode.ICKSTK);
	adjust(lv, n);
	if (Debug) do
		for (i=pb, St, SYMREC) dumpsym(i, 1);
	end
	xrparen();
	rv := stmt();
	cleanup();
	if (\rv) do
		gen1(tcode.INUM, 0);
		gen0(tcode.IPOP);
	end
	gen1(tcode.ICLAB, Exitlab);
	if (pub) gen0(tcode.IECTX);
	gen0(tcode.IEND);
	Exitlab := 0;
	Method_adjust := 0;
end


procdecl() do var k, flgs;
	Token := scan();
	flgs := PROC|PROTO;
	while (1) do
		if (forcesym(0)) return 0;
		k := newsym(Text, 0, 0, flgs);
		Token := scan();
		xlparen();
		Symbols[k+SSIZE] := constval();
		Symbols[k+SVAL] := newlab();
		xrparen();
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xsemi();
end


allocvar(k, local) ie (local) do
			Symbols[k+SVAL] := Locladdr;
			if (Symbols[k+SSIZE])
				Symbols[k+SVAL] :=
					Symbols[k+SVAL] +
					Symbols[k+SSIZE]-1;
			Locladdr := Locladdr +
				(Symbols[k+SSIZE]->
				Symbols[k+SSIZE]: 1);
		end
		else ie (Classctx \= %1) do
			Symbols[k+SVAL] := Classoff;
			Classoff := Classoff +
				(Symbols[k+SSIZE]->
				Symbols[k+SSIZE]: 1);
		end
		else do
			Symbols[k+SVAL] := newlab();
			gen1(tcode.IDLAB, Symbols[k+SVAL]);
			gen1(tcode.IDECL, Symbols[k+SSIZE]);
		end


vardecl(pub, local) do var n, tok, k;
	tok := Token;
	Token := scan();
	while (1) do
		if (forcesym(0)) return 0;
		k := newsym(Text, 0, 0,
			(tok=KCONST-> CNST: local-> 0 : GLOBL) |
			(pub-> PUBLC: 0));
		Token := scan();
		ie (tok = KCONST) do
			equsign();
			Symbols[k+SVAL] := constval();
		end
		else do
			if (Token = LBRACK \/ Token = BYTEOP) do
				tok := Token;
				Token := scan();
				n := constval();
				ie (tok = BYTEOP)
					n := n+1 >> 1;
				else
					xrbrack();
				if (n < 1) error(packed"bad vector size",
					Symbols[k+SNAME]);
				Symbols[k+SSIZE] := n;
			end
			allocvar(k, local);
			if (Debug) dumpsym(k, local);
		end
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xsemi();
end


defstruct(pub) do var base, count;
	Token := scan();
	if (forcesym(packed"struct name expected")) return 0;
	base := newsym(Text, 0, 0, CNST | (pub-> PUBLC: 0));
	Token := scan();
	equsign();
	count := 0;
	while (1) do
		if (forcesym(packed"member name expected")) return 0;
		newsym(Text, 0, count, CNST | (pub-> PUBLC: 0));
		Token := scan();
		count := count+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	Symbols[base+SVAL] := count;
	xsemi();
end


objdef(local) do var k, i, cl;
	Token := scan();
	while (1) do
		if (forcesym(0)) return 0;
		k := newsym(Text, 0, 0, OBJCT | (local->0:GLOBL));
		Token := scan();
		match(LBRACK, packed"missing '['");
		if (forcesym(0)) return 0;
		cl := findclassref(Text, 1);
		Token := scan();
		Symbols[k+SSIZE] := cl=%1-> %1: Libcache[cl+SSIZE];
		allocvar(k, local);
		Symbols[k+SSIZE] := cl=%1-> %1: cl;
		xrbrack();
		if (cl = %1) St := St-SYMREC;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xsemi();
end


ifacedecl() do var k, slot;
	Token := scan();
	while (1) do
		if (forcesym(0)) return 0;
		Token := scan();
		xlparen();
		constval();
		xrparen();
		if (Token = BINOP /\ Op = Sym_equ) do
			Token := scan();
			slot := constval();
			if (slot < Bcslot)
				error(packed"slot already in use", 0);
			Bcslot := slot;
		end
		if (Bcslot >= 19)
			error(packed"extension procedure not available", 0);
		Bcslot := Bcslot+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xsemi();
end


accept() while (Token = KVAR \/ Token = KCONST \/ Token = SYMBOL \/
		Token = KDECL \/ Token = KSTRUCT \/ Token = KIFACE \/
		Token = KPUBLIC \/ Token = KCLASS \/ Token = KOBJECT \/
		Token = KMODULE \/ Token = METAOP
	)
		declaration(0);


redirlist() do var b;
	xlparen();
	while (Token = SYMBOL) do
		b := findclass(Text, 1);
		if (b \= %1) newsym(packed"+", 0, b, REDIR);
		Token := scan();
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xrparen();
end


classdecl(pub) do
	var	cl;
	var	p, i, k;
	var	loc_nt, loc_st;
	var	thiscl::TEXTLEN+1;

	if (pub /\ \Modname::0)
		error(packed"public class in anonymous module", 0);
	if (Classctx \= %1) error(packed"nested class", 0);
	Token := scan();
	if (forcesym(0)) return 0;
	strcpy(thiscl, Text);
	Token := scan();
	loc_st := St;
	redirlist();
	loc_nt := Nt;
	cl := newsym(thiscl, 0, %1, CLSS | (pub->PUBLC:0));
	Classctx := cl;
	Classoff := 0;
	Symbols[Selfobj+SSIZE] := Classctx;
	accept();
	Classctx := %1;
	Symbols[cl+SSIZE] := Classoff = 0-> 1: Classoff;
	match(KEND, packed"missing END");
	newsym(Modname::0-> Modname: packed"**", 0, 0, PUBLC);
	rptdecls(cl, St);
	for (p=cl, St, SYMREC) do
		if (p = cl \/ Symbols[p+SFLAGS] & PUBLC = PUBLC) do
			k := length(Symbols[p+SNAME])+1;
			t.memcopy(@Names::loc_nt, Symbols[p+SNAME], k);
			for (i=0, SYMREC) Libcache[Lt+i] := Symbols[p+i];
			Libcache[Lt+SNAME] := @Names::loc_nt;
			loc_nt := loc_nt+k;
			Lt := Lt+SYMREC;
		end
	end
	Libcache[Lt-SYMREC+SFLAGS] := 0;
	Nt := loc_nt;
	St := loc_st;
end


moddecl() do var k, i, j;
	Token := scan();
	if (forcesym(packed"module name expected")) return 0;
	if (Modname::0) error(packed"module redefinition", Text);
	k := length(Text);
	Modname::0 := '*';
	strcpy(@Modname::1, Text);
	Modname::(k+1) := '*';
	Modname::(k+2) := 0;
	Token := scan();
	redirlist();
	Modctx := St;
	xsemi();
	for (i=0, Lt, SYMREC) do
		if (Libcache[i+SFLAGS] & CLSS) k := i;
		if (	\Libcache[i+SFLAGS] /\
			strequ(Libcache[i+SNAME], Modname)
		) do
			if (Lt-i-SYMREC)
				ie (TUNE16) do
					t.memcopy(@Libcache[k],
						@Libcache[i+SYMREC],
						(Lt-i-SYMREC)*2);
				end
				else do
					for (j=0, (Lt-i-SYMREC))
						Libcache[k+j] :=
							Libcache[i+SYMREC+j];
				end
			Lt := Lt-(i-k+SYMREC);
			i := k;
		end
	end
end


declaration(local) do var pub;
	pub:= 0;
	if (Token = KPUBLIC) do
		pub := 1;
		Token := scan();
		if (	(Token \= SYMBOL /\ Token \= KSTRUCT /\
			 Token \= KCONST \/ Classctx = %1) /\
			Token \= KCLASS
		)
			error(packed"'PUBLIC' in wrong context", 0);
	end
	ie (Token = KVAR \/ Token = KCONST) do
		vardecl(pub, local);
	end
	else ie (Token = KDECL) do
		procdecl();
	end
	else ie (Token = KSTRUCT) do
		defstruct(pub);
	end
	else ie (Token = SYMBOL) do
		procdef(pub);
	end
	else ie (Token = KOBJECT) do
		objdef(local);
	end
	else ie (Token = KCLASS) do
		classdecl(pub);
	end
	else ie (Token = KIFACE) do
		ifacedecl();
	end
	else ie (Token = KMODULE) do
		if (Classctx \= %1)
			error(packed"'MODULE' in wrong context", 0);
		moddecl();
	end
	else ie (Token = METAOP) do
		metacmd();
	end
	else do
		error(packed"bad declaration", 0);
		synch(0);
	end
end


initfuncs() do
	Selfobj := newsym(packed"self", 0, 0, OBJCT|GLOBL);
	newsym(packed"_version_", 0, 30604, CNST);
	Sym_equ := findop2(packed"=");
	Sym_sub := findop2(packed"-");
	Sym_mod := findop2(packed"mod");
	Sym_add := findop2(packed"+");
	Sym_mul := findop2(packed"*");
	Sym_bar := findop2(packed"|");
	Sym_not := findop2(packed"~");
end


do
	initvars();
	initfuncs();
	rdclsdir(packed"CLASSDIR", 0);
	gen2(tcode.IINIT, 3, 1);
	gen1(tcode.ICALL, 1);
	gen1(tcode.IHALT, 0);
	gen1(tcode.IDATA, 0);
	Token := scan();
	accept();
	if (Token \= KDO) fatal(packed"'DO' or declaration expected", 0);
	gen1(tcode.ICLAB, 1);
	gen0(tcode.IHDR);
	compound();
	if (Token \= ENDFILE) error(packed"trailing declarations", 0);
	rptdecls(0, St);
	gen0(tcode.IEND);
	commit();
	flush();
	if (\Errcount) wrclsdir();
	if (Errcount) t.close(t.open(packed"T3X.ERR", T3X.OWRITE));
end

