/*
 *	TXX -- a virtual Tcode3 machine
 *	Copyright (C) 1997-2000 Nils M Holm. All rights reserved.
 *	See the file LICENSE for conditions of use.
 *
 *	SYNTAX: TXX PROGRAM [OPTIONS]
 */

#include <stdio.h>

#ifdef dos
 #ifdef __TURBOC__
  #include <alloc.h>
  #include <io.h>
 #endif
#endif

#define MAIN
#include "txx.h"
#include "tcode.h"

void	ifail();

void	iadd(), iband(), ibnot(), ibor(), ibrf(), ibrt(), ibshl(),
	ibshr(), ibxor(), icall(), icalr(), icalx(), icctx(), ickstk(),
	iclean(), ideref(), idiv(), idnext(), idrefb(), idup(), iectx(),
	iend(), iequ(), isys(), igrtr(), igteq(), ihalt(), ihdr(),
	iincg(), iinci(), iincl(), iind(), iindb(), ijump(), ildg(),
	ildgv(), ildi(), ildiv(), ildl(), ildlab(), ildlv(), iless(),
	iline(), ilnot(), ilteq(), imod(), imul(), inbrf(), inbrt(),
	ineg(), inequ(), inop(), inum(), ipop(), isavg(), isavi(),
	isavl(), iself(), istack(), istorb(), istore(), isub(),
	iswap(), iudiv(), iugrtr(), iugteq(), iuless(), iulteq(),
	iumul(), iunext();

void	(*Builtins[])() = {
	ifail, ifail, ifail, ifail, ifail, ifail, ifail,
	ifail, ifail, ifail,
	ihdr, iend, icctx, iectx, ipop, idup, iswap, istack, iclean,
	ineg, ilnot, ibnot, inop, iincg, iinci, iincl, imul, idiv,
	iumul, iudiv, imod, iadd, isub, iband, ibor, ibxor, ibshl,
	ibshr, iequ, inequ, iless, igrtr, ilteq, igteq, iuless,
	iugrtr, iulteq, iugteq, ildg, ildgv, ildl, ildlv, ildi,
	ildiv, ildlab, inum, iself, iind, iindb, ideref, idrefb,
	isavg, isavl, isavi, istore, istorb, icall, icalr, isys,
	ibrf, ibrt, inbrf, inbrt, ijump, iunext, idnext, ihalt,
	ifail, ifail, icalx, iline, ifail, ifail, ifail, ickstk
};

#ifdef dos
 #define TEXTSIZE	64000U
 #define DATASIZE	TEXTSIZE
#else
 #ifdef SMALL
  #define TEXTSIZE	20480
  #define DATASIZE	28672
 #else
  #define TEXTSIZE	65536L
  #define DATASIZE	TEXTSIZE
 #endif
#endif

#ifdef SMALL
 #define NLABELS	1024
 #define NMARKS		512
#else
 #define NLABELS	4096
 #define NMARKS		4096
#endif

#define LDBUFLEN	1024

struct mark {
	int	l, t, d;
};

/* shared variables are defined in 'txx.h' */
uchar		*Text;
CELL		*Wdata;
UCELL		*Uwdata;
UCELL		Etext, Edata;
int		Tags[NLABELS];
struct mark	Marks[NMARKS];
uint		Mt;
uint		Dptr, Tptr;
uint		Op;
CELL		A1, A2;
UCELL		U1;
UCELL		Ip, Fp;
UCELL		Self;
int		Execmax;

uchar		Ldbuf[LDBUFLEN];
int		Lfd, Lp, Lz;

char	*Opcodes[] = {
		"OOPS!", "CLAB", "DLAB", "DECL", "DATA", "CREF",
		"DREF", "STR", "PSTR", "INIT", "HDR", "END",
		"CCTX", "ECTX", "POP", "DUP", "SWAP", "STACK",
		"CLEAN", "NEG", "LNOT", "BNOT", "NOP", "INCG",
		"INCI", "INCL", "MUL", "DIV", "UMUL", "UDIV",
		"MOD", "ADD", "SUB", "BAND", "BOR", "BXOR",
		"BSHL", "BSHR", "EQU", "NEQU", "LESS", "GRTR",
		"LTEQ", "GTEQ", "ULESS", "UGRTR", "ULTEQ", "UGTEQ",
		"LDG", "LDGV", "LDL", "LDLV", "LDI", "LDIV",
		"LDLAB", "NUM", "SELF", "IND", "INDB", "DEREF",
		"DREFB", "SAVG", "SAVL", "SAVI", "STORE", "STORB",
		"CALL", "CALR", "SYS", "BRF", "BRT", "NBRF",
		"NBRT", "JUMP", "UNEXT", "DNEXT", "HALT", "PUB",
		"EXT", "CALX", "LINE", "GSYM", "LSYM", "ISYM",
		"CKSTK"
};


#ifdef __TURBOC__
void fail(char *m, char *n);
#endif

void fail(m, n)
char	*m, *n;
{
	if (n)
		fprintf(stderr, "TXX core fault: %s: %s\n", m, n);
	else
		fprintf(stderr, "TXX core fault: %s\n", m);
	fprintf(stderr, "IP=%u, SP=%u, FP=%u, RR=%d\n"
			"Slot#=%d, Etext=%u, Edata=%u\n",
		Ip, Sp<<1, Fp<<1, Acc, Execmax, Etext, Edata);
	if (Op) {
		fprintf(stderr, "Code: ");
		if (Op & 128) {
			if (Op == IINCL || Op == IINCG || Op == IINCI)
				fprintf(stderr, "%s %d (%u) %d\n",
					Opcodes[Op&127], A1, U1, A2);
			else
				fprintf(stderr, "%s %d (%u)\n",
					Opcodes[Op&127], A1, U1);
		}
		else {
			fprintf(stderr, "%s\n", Opcodes[Op]);
		}
	}
	exit(2);
}


#ifdef dos

/* Some rock'n'roll required for pointer passing to external procedures
 * under DOS. This malloc() routine always delivers blocks which begin
 * at a segment boundary.
 * NEVER attempt to free() a block returned by this function!
 */
void *_malloc(n)
uint	n;
{
	void	*x;

	if ((x = malloc(n+16)) == NULL) return NULL;
	return MK_FP(FP_SEG(x)+1, FP_OFF(x)&~0xf);
}

#define malloc	_malloc

#endif


void init() {
	int	i;

	Text = (uchar *) malloc(TEXTSIZE);
	Data = (uchar *) malloc(DATASIZE);
	if (!Text || !Data) fail("insufficient memory", 0);
	Wdata = (CELL *) Data;
	Stack = Wdata;
	Uwdata = (UCELL *) Data;
	Ustack = Uwdata;
	Sp = Fp = DATASIZE/2;
	Etext = 0;
	Edata = 0;
	for (i=0; i<NLABELS; i++) Tags[i] = -1;
	Mt = 0;
	Tptr = 0;
	Dptr = 0;
	A1 = 0;
	A2 = 0;
	Op = 0;
	Self = 0;
	Execmax = sizeof(Primitives) / sizeof(Primitives[0]);
	libinit();
}


void addlabel(op, id) {
	uint	a, i, o;
	uint	here;

	if (id >= NLABELS) fail("label too big", 0);
	if (Tags[id] != -1) fail("duplicate label", 0);
	Tags[id] = op == ICLAB? Tptr: Dptr;
	here = op == IDLAB? Dptr: Tptr;
	for (i=0; i<Mt; i++) if (Marks[i].l == id) break;
	if (i >= Mt) return;
	a = Marks[i].t;
	while (a) {
		o = a;
		a = (Text[a+1] << 8) | Text[a];
		Text[o] = here & 0xff;
		Text[o+1] = here >> 8;
	}
	a = Marks[i].d;
	while (a) {
		o = a;
		a = (Data[a+1] << 8) | Data[a];
		Data[o] = here & 0xff;
		Data[o+1] = here >> 8;
	}
}


int addmark(id, dref) {
	int	i, old;

	for (i=0; i<Mt; i++) {
		if (Marks[i].l == id) {
			if (dref) {
				old = Marks[i].d;
				Marks[i].d = Dptr;
			}
			else {
				old = Marks[i].t;
				Marks[i].t = Tptr;
			}
			return(old);
		}
	}
	if (Mt >= NMARKS) fail("too many marks", 0);
	Marks[Mt].l = id;
	if (dref) {
		Marks[Mt].d = Dptr;
		Marks[Mt].t = 0;
	}
	else {
		Marks[Mt].t = Tptr;
		Marks[Mt].d = 0;
	}
	Mt++;
	return(0);
}


int findlab(id, dref) {
	if (Tags[id] != -1) return(Tags[id]);
	return(addmark(id, dref));
}


uchar rdc() {
	if (Lp >= Lz) {
		Lz = read(Lfd, Ldbuf, LDBUFLEN);
		Lp = 0;
	}
	return(Ldbuf[Lp++]);
}


uint rdop() {
	uint	n = rdc();
	return((rdc() << 8) | n);
}


skip(n) {
	int	i;

	for (i=0; i<n; i++) rdc();
}


copy(n, p)
int	n;
uchar	*p;
{
	int	i;

	for (i=0; i<n; i++) p[i] = rdc();
}


dgrow(n) {
	if ((Dptr += n) > DATASIZE) fail("data segment overflow", 0);
}


tgrow(n) {
	if ((Tptr += n) > TEXTSIZE) fail("text segment overflow", 0);
}


int resolve() {
	UCELL	i, op, arg, arg2, a2f;
	uint	b;

	op = rdc();
	if (op == '#') {
		while (Ldbuf[Lp++] != '\n')
			if (Lp >= LDBUFLEN-1)
				fail("#! header too long", 0);
	}
	else {
		Lp--;
	}
	if (Ldbuf[Lp] != IINIT)
		fail("bad input program (magic match failed)", 0);
	for (;;) {
		op = rdc();
		if (Lz < 1) break;
		if (op & 128) arg = rdop();
		a2f = 0;
		if (op==IINCG || op==IINCL || op==IINCI || op==IINIT) {
			arg2 = rdop();
			a2f = 1;
		}
		switch (op) {
		case ICLAB:
		case IDLAB:
			addlabel(op, arg);
			op = 0;
			break;

		case IDECL:
			dgrow(arg? (arg<<1): 2);
			op = 0;
			break;

		case IDATA:
			Wdata[Dptr>>1] = arg;
			dgrow(2);
			op = 0;
			break;

		case ICREF:
		case IDREF:
			Wdata[Dptr>>1] = findlab(arg, 1);
			dgrow(2);
			op = 0;
			break;

		case IPSTR:
			b = Dptr;
			dgrow(arg);
			copy(arg, &Data[b]);
			do { Data[Dptr++] = 0; } while (Dptr & 1);
			op = 0;
			break;

		case ISTR:
			b = Dptr;
			dgrow(arg*2);
			for (i=0; i<arg; i++) {
				Data[b++] = rdc();
				Data[b++] = 0;
			}
			do { Data[Dptr++] = 0; } while (Dptr & 1);
			op = 0;
			break;

		case IINCG: case ILDG: case ILDGV: case ISAVG:
		case ILDLAB: case ICALL: case IBRF: case IBRT:
		case INBRF: case INBRT: case IJUMP: case IUNEXT:
		case IDNEXT:
			Text[Tptr] = op;
			tgrow(1);
			arg = findlab(arg, 0);
			break;

		case IPUB:
			arg = rdop();
			skip(arg+2);
			op = 0;
			break;

		case ILSYM:
		case IGSYM:
		case IISYM:
			arg = rdop();
			skip(arg+2);
			op = 0;
			break;

		case IEXT:
			arg = rdop();
			copy(arg, Text);
			Text[arg] = 0;
			fail("external reference", (char *) Text);
			break;

		case IINIT:
			if (arg == 1 || arg == 2)
				fail("obsolete Tcode version (1,2)", NULL);
			else if (arg == 3)
				;
			else if (arg == (3 | 256))
				return -1;
			else
				fail("unsupported tcode version", NULL);
			op = 0;
			break;

		default:
			Text[Tptr] = op;
			tgrow(1);
		}
		if (op & 128) {
			Text[Tptr] = arg & 0xff;
			Text[Tptr+1] = arg >> 8;
			tgrow(2);
			if (a2f) {
				Text[Tptr] = arg2 & 0xff;
				Text[Tptr+1] = arg2 >> 8;
				tgrow(2);
			}
		}
	}
	Etext = Tptr;
	Edata = Dptr;
}


int load(name)
char	*name;
{
	int	r;

	if ((Lfd = open(name, 0)) < 0) fail("cannot open program", name);
	Lp = Lz = 0;
	r = resolve();
	close(Lfd);
	return r;
}


loadimage(name)
char	*name;
{
	uchar	hd[5];
	uint	tlen;
	int	i;

	if ((Lfd = open(name, 0)) < 0) fail("cannot open program", name);
	if (read(Lfd, hd, 5) != 5) fail("short file -- oops!", name);
	tlen = (hd[4] << 8) | hd[3];
	for (i=0; i<5; i++) Text[i] = INOP;
	tlen -= 5;
	if (read(Lfd, &Text[5], tlen) != tlen) fail("file read error", name);
	if (read(Lfd, Data, DATASIZE) == -1) fail("file read error", name);
	close(Lfd);
}


void primitive(id) {
	if (id < 0 || id >= Execmax-1 || !Primitives[id])
		fail("undefined primitive", 0);
	(*Primitives[id])();
}


void divz() { fail("divide by zero", 0); }

void iadd() { Stack[Sp+1] += Stack[Sp]; Sp++; }
void iband() { Stack[Sp+1] &= Stack[Sp]; Sp++; }
void ibnot() { Stack[Sp] = ~Stack[Sp]; }
void ibor() { Stack[Sp+1] |= Stack[Sp]; Sp++; }
void ibrf() { if (!Stack[Sp++]) Ip = U1; }
void ibrt() { if (Stack[Sp++]) Ip = U1; }
void ibshl() { Stack[Sp+1] <<= Stack[Sp]; Sp++; }
void ibshr() { Ustack[Sp+1] >>= Ustack[Sp]; Sp++; }
void ibxor() { Stack[Sp+1] ^= Stack[Sp]; Sp++; }
void icall() { Stack[--Sp] = Ip; Ip = U1; }
void icalr() { UCELL t; t = Stack[Sp]; Stack[Sp] = Ip; Ip = t; }
void icalx() { fail("call to external procedure", 0); }
void icctx() { Stack[--Sp] = Self; Self = Stack[Sp+3]; }
void iclean() { Sp += A1; Stack[--Sp] = Acc; }
void ideref() { Stack[Sp+1] += (Stack[Sp]<<1); Sp++; }
void idiv() { if (!Stack[Sp]) divz(); Stack[Sp+1] /= Stack[Sp]; Sp++; }
void idnext() { if (Stack[Sp+1] <= Stack[Sp]) Ip = U1; Sp+=2; }
void idrefb() { Stack[Sp+1] += Stack[Sp]; Sp++; }
void idup() { Stack[Sp-1] = Stack[Sp]; Sp--; }
void iectx() { Self = Stack[Sp++]; }
void iend() { Fp = Stack[Sp++]; Ip = Stack[Sp++]; }
void iequ() { Stack[Sp+1] = Stack[Sp+1] == Stack[Sp]? -1: 0; Sp++; }
void isys() { primitive(U1); }
void ifail() { Op = 0; fail("illegal opcode", 0); }
void igrtr() { Stack[Sp+1] = Stack[Sp+1] > Stack[Sp]? -1: 0; Sp++; }
void igteq() { Stack[Sp+1] = Stack[Sp+1] >= Stack[Sp]? -1: 0; Sp++; }
void ihalt() { exit(A1); }
void ihdr() { Stack[--Sp] = Fp; Fp = Sp; }
void iincg() { A2 = (Text[Ip+1] << 8) | Text[Ip]; Ip += 2;
	Wdata[U1>>1] += A2; }
void iinci() { A2 = (Text[Ip+1] << 8) | Text[Ip]; Ip += 2;
	Wdata[(Self>>1)+U1] += A2; }
void iincl() { A2 = (Text[Ip+1] << 8) | Text[Ip]; Ip += 2;
	Stack[Fp-A1] += A2; }
void iind() { Stack[Sp] = Wdata[Ustack[Sp]>>1]; }
void iindb() { Stack[Sp] = Data[Ustack[Sp]]; }
void ijump() { Ip = U1; }
void ildg() { Stack[--Sp] = Wdata[U1>>1]; }
void ildgv() { Stack[--Sp] = U1; }
void ildi() { Stack[--Sp] = Wdata[(Self>>1) + U1]; }
void ildiv() { Stack[--Sp] = Self + (U1<<1); }
void ildl() { Stack[--Sp] = Stack[Fp-A1]; }
void ildlab() { Stack[--Sp] = U1; }
void ildlv() { Stack[--Sp] = (Fp-A1)<<1; }
void iless() { Stack[Sp+1] = Stack[Sp+1] < Stack[Sp]? -1: 0; Sp++; }
void iline() {}
void ilnot() { Stack[Sp] = Stack[Sp]? 0: -1; }
void ilteq() { Stack[Sp+1] = Stack[Sp+1] <= Stack[Sp]? -1: 0; Sp++; }
void imod() { if (!Stack[Sp]) divz(); Ustack[Sp+1] %= Ustack[Sp]; Sp++; }
void imul() { Stack[Sp+1] *= Stack[Sp]; Sp++; }
void inbrf() { if (!Stack[Sp]) Ip = U1; }
void inbrt() { if (Stack[Sp]) Ip = U1; }
void ineg() { Stack[Sp] = -Stack[Sp]; }
void inequ() { Stack[Sp+1] = Stack[Sp+1] != Stack[Sp]? -1: 0; Sp++; }
void inop() { }
void inum() { Stack[--Sp] = A1; }
void ipop() { Acc = Stack[Sp++]; }
void isavg() { Wdata[U1>>1] = Stack[Sp++]; }
void isavi() { Wdata[(Self>>1)+U1] = Stack[Sp++]; }
void isavl() { Stack[Fp-A1] = Stack[Sp++]; }
void iself() { Stack[--Sp] = Self; }
void istack() { Sp -= A1; }
void istorb() { Data[Ustack[Sp+1]] = Ustack[Sp]; Sp+=2; }
void istore() { Wdata[Ustack[Sp+1]>>1] = Stack[Sp]; Sp+=2; }
void isub() { Stack[Sp+1] -= Stack[Sp]; Sp++; }
void iswap() { UCELL t; t = Stack[Sp+1]; Stack[Sp+1] = Stack[Sp];
	Stack[Sp] = t; }
void iudiv() { if (!Stack[Sp]) divz(); Ustack[Sp+1] /= Ustack[Sp]; Sp++; }
void iugrtr() { Ustack[Sp+1] = Ustack[Sp+1] > Ustack[Sp]? -1: 0; Sp++; }
void iugteq() { Ustack[Sp+1] = Ustack[Sp+1] >= Ustack[Sp]? -1: 0; Sp++; }
void iuless() { Ustack[Sp+1] = Ustack[Sp+1] < Ustack[Sp]? -1: 0; Sp++; }
void iulteq() { Ustack[Sp+1] = Ustack[Sp+1] <= Ustack[Sp]? -1: 0; Sp++; }
void iumul() { Ustack[Sp+1] *= Ustack[Sp]; Sp++; }
void iunext() { if (Stack[Sp+1] >= Stack[Sp]) Ip = U1; Sp+=2; }
void ickstk() { if ((Sp<<1) < Edata) fail("stack overflow", 0); }


run() {
	uint	o;

	for (Ip = 0;;) {
		Op = Text[Ip++];
		if (Op & 128) {
			A1 = (Text[Ip+1] << 8) | Text[Ip];
			Ip += 2;
		}
		o = Op & ~128;
		if (o >= IENDOFSET) ifail();
		U1 = A1;
		(*Builtins[o])();
	}
}


void version() {
	printf("TXX 3.3 -- Copyright (C) 1997-2000 Nils M Holm\n");
	printf("Tcode runtime engine V1.3 (Tcode3)");
#ifdef SHL_SUPPORT
	printf(" with shared library support");
#endif
	putchar('\n');
	exit(0);
}


main(argc, argv)
int	argc;
char	**argv;
{
	Argcount = argc-1;
	Argvec = argv+1;
	init();
	if (argc < 2) version();
	if (load(argv[1]) < 0) loadimage(argv[1]);
	run();
	return(0);
}

