!	TXPRINT -- Pretty-print T3X programs
!	Copyright (C) 2000 Nils M Holm.  All rights reserved.
!	See the file LICENSE for conditions of use.

#classpath "../classes";
#packstrings;

module txprint(t3x, string, iostream, memory, char);

object	t[t3x], str[string], chr[char], mem[memory];

const	BUFLEN	= 1024;
const	TEXTLEN = 129;
const	POOLSIZE = 2048;

const	ENDFILE=-1, OPERATOR=1, KEYWORD=2, SYMBOL=3, NUMBER=4, STRLIT=5,
	COMMENT=6;

struct	OPLEN = OLEN, ONAME, OTOK, OFLGS;

struct	CONV = CCASE, CPREFIX, CSUFFIX;

const	NONE = 0, LCASE = 1, UCASE = 2, CAPITAL = 3;

const	LPAD =	0x0001,	! pad left side
	RPAD =	0x0002,	! pad right side
	PAD = LPAD|RPAD,
	SID =	0x0004,	! start indentation
	EID =	0x0008,	! end indentation
	LBRK =	0x0010,	! line break before
	RBRK =	0x0020,	! line break after
	BLNK =	0x0040,	! blank line before
	STCK =	0x0080,	! stick right
	BLNI =	0x0100,	! BLNK if no indent
	RSHP =	0x0200,	! righthand side half pad
	LSHP =	0x0400,	! lefthand side half pad
	PTAB =	0x0800,	! pad with tabs
	PCOM =	0x1000,	! pending comment
	STID =	0x2000,	! start temp ident
	ETID =	0x4000;	! end temp ident


object	In[iostream], Out[iostream];
var	Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Text::TEXTLEN+1;
var	Tlength;
var	Flags;
var	Ops;
var	Kwds;
var	Rejected;
var	Line;
var	O_html, O_adjust;
var	Convtab;
var	Prologue, Epilogue;
var	Pool::POOLSIZE;
var	Rulefile;
var	Dopad, Donepad, Dobrk, Donebrk, Dotblbrk,
	Doneblnk, Dostick, Donecmnt;
var	Halfpad;
var	Indent, Tmpind;
var	Xsize;
var	Margin;
var	Tabsize;
var	Foundlb, Foundrb, Intbl;


init() do
	chr.init();
	mem.init(Pool, POOLSIZE);
	In.create(0, Inbuf, BUFLEN, iostream.FREAD);
	Out.create(1, Outbuf, BUFLEN, iostream.FWRITE);
	Tabsize := 8;
	Margin := 79;
	Prologue := 0;
	Epilogue := 0;
	Rulefile := 0;
	O_html := 0;
	O_adjust := 0;
	Ops := [
		[ 1, "+",	OPERATOR,	0		],
		[ 1, "*",	OPERATOR,	0		],
		[ 1, ";",	OPERATOR,	RBRK|ETID	],
		[ 1, ",",	OPERATOR,	RPAD		],
		[ 1, "(",	OPERATOR,	0		],
		[ 1, ")",	OPERATOR,	RSHP		],
		[ 1, "[",	OPERATOR,	0		],
		[ 1, "]",	OPERATOR,	0		],
		[ 1, "=",	OPERATOR,	PAD		],
		[ 1, "&",	OPERATOR,	0		],
		[ 1, "|",	OPERATOR,	0		],
		[ 1, "^",	OPERATOR,	0		],
		[ 1, "@",	OPERATOR,	0		],
		[ 1, "~",	OPERATOR,	0		],
		[ 1, ":",	OPERATOR,	RPAD		],
		[ 2, "::",	OPERATOR,	0		],
		[ 2, ":=",	OPERATOR,	PAD		],
		[ 1, "\\",	OPERATOR,	0		],
		[ 2, "\\/",	OPERATOR,	PAD		],
		[ 2, "\\=",	OPERATOR,	PAD		],
		[ 1, "<",	OPERATOR,	0		],
		[ 2, "<<",	OPERATOR,	0		],
		[ 2, "<=",	OPERATOR,	PAD		],
		[ 1, ">",	OPERATOR,	PAD		],
		[ 2, ">>",	OPERATOR,	0		],
		[ 2, ">=",	OPERATOR,	PAD		],
		[ 1, "-",	OPERATOR,	0		],
		[ 2, "->",	OPERATOR,	RPAD		],
		[ 1, "/",	OPERATOR,	0		],
		[ 2, "/\\",	OPERATOR,	PAD		],
		[ 1, ".",	OPERATOR,	0		],
		[ 2, "./",	OPERATOR,	0		],
		[ 2, ".*",	OPERATOR,	0		],
		[ 2, ".<",	OPERATOR,	PAD		],
		[ 3, ".<=",	OPERATOR,	PAD		],
		[ 2, ".>",	OPERATOR,	PAD		],
		[ 3, ".>=",	OPERATOR,	PAD		],
		[ 1, "#",	OPERATOR,	LBRK		],
		[ 3, "mod",	KEYWORD,	PAD		],
		[ 0, 0,		0,		0		]
	];
	Kwds := [
		[ "call",	PAD			],
		[ "class",	PAD|BLNK		],
		[ "const",	PAD|LBRK|PTAB|STID	],
		[ "decl",	PAD|LBRK|PTAB|STID	],
		[ "do",		PAD|RBRK|SID|BLNI	],
		[ "else",	PAD|LBRK		],
		[ "end",	LBRK|RBRK|EID		],
		[ "for",	PAD			],
		[ "halt",	PAD			],
		[ "ie",		PAD			],
		[ "if",		PAD			],
		[ "interface",	PAD|LBRK|PTAB|STID	],
		[ "leave",	PAD			],
		[ "loop",	PAD			],
		[ "mod",	PAD			],
		[ "module",	PAD|BLNK		],
		[ "object",	PAD|LBRK|PTAB|STID	],
		[ "packed",	LPAD			],
		[ "public",	PAD|BLNK|STCK		],
		[ "return",	PAD			],
		[ "send",	PAD			],
		[ "struct",	PAD|LBRK|PTAB|STID	],
		[ "var",	PAD|LBRK|PTAB|STID	],
		[ "while",	PAD			],
		0
	];
	Convtab := [
		[ NONE,	0,	0	],	! VOID
		[ NONE,	0,	0	],	! Operators
		[ NONE,	0,	0	],	! Keywords
		[ NONE,	0,	0	],	! Symbols
		[ NONE,	0,	0	],	! Numbers, Chars
		[ NONE,	0,	0	],	! Strings
		[ NONE,	0,	0	],	! Comments
	];
	Rejected := 0;
	Line := 1;
	Dopad := 0;
	Donepad := 0;
	Dobrk := 0;
	Donebrk := 1;
	Dotblbrk := 0;
	Doneblnk := 1;
	Donecmnt := 0;
	Dostick := 0;
	Halfpad := 0;
	Foundlb := 0;
	Foundrb := 0;
	Intbl := 0;
	Indent := 0;
	Tmpind := 0;
end


_mwrites(s) t.write(2, s, str.length(s));


mwrites(s) do var b::5;
	_mwrites(s);
	_mwrites(t.newline(b));
end


error(m, n) do var b::100;
	_mwrites(str.format(b, "TXPRINT: %D: ", [(Line)]));
	mwrites(str.format(b, m, [(n)]));
	halt 1;
end


toolong() error("token too long", 0);


putc(c) do
	ie (\O_html) Out.wrch(c);
	else ie (c = '<') Out.writes("&lt;");
	else ie (c = '>') Out.writes("&gt;");
	else ie (c = '&') Out.writes("&amp;");
	else Out.wrch(c);
end


putstr(s) do
	var	b::TEXTLEN+5, i, j, k, c;

	if (\O_html) do
		Out.writes(s);
		return;
	end
	k := str.length(s);
	j := 0;
	for (i=0, k) do
		c := s::i;
		if (j >= TEXTLEN) do
			b::j := 0;
			Out.writes(b);
			j := 0;
		end
		ie (c = '<') do
			str.copy(@b::j, "&lt;"); j := j+4;
		end
		else ie (c = '>') do
			str.copy(@b::j, "&gt;"); j := j+4;
		end
		else ie (c = '&') do
			str.copy(@b::j, "&amp;"); j := j+5;
		end
		else do
			b::j := c; j := j+1;
		end
	end
	b::j := 0;
	if (j) Out.writes(b);
end


getc() do var c;
	if (Rejected) do
		c := Rejected;
		Rejected := 0;
		return c;
	end
	return In.rdch();
end


getce(b) do
	b::0 := getc();
	b::1 := 0;
	if (b::0 \= '\\') return b::0;
	b::1 := getc();
	b::2 := 0;
	return b::0;
end


strequ(a, b) return \str.comp(a, b);


lowcase(s) do var i;
	i := 0; while (s::i) do
		s::i := chr.lcase(s::i);
		i := i+1;
	end
	return i;
end


upcase(s) do var i;
	i := 0; while (s::i) do
		s::i := chr.ucase(s::i);
		i := i+1;
	end
	return i;
end


capitalize(s) do var i;
	i := lowcase(s);
	s::0 := chr.ucase(s::0);
	return i;
end


findkw(s) do var b::TEXTLEN, i, c;
	str.copy(b, s);
	lowcase(b);
	c := b::0;
	for (i=0, 128) do
		if (\Kwds[i]) leave;
		if (c = Kwds[i][0]::0 /\ strequ(b, Kwds[i][0]))
			leave;
	end
	if (\Kwds[i]) return 0;
	Flags := Kwds[i][1];
	return KEYWORD;
end


findop(c) do var nmatch, i, pos;
	nmatch := 0;
	i := 0;
	pos := %1;
	while (Ops[i][OLEN] > nmatch) do
		if (c = Ops[i][ONAME]::nmatch) do
			c := getc();
			pos := i;
			nmatch := nmatch+1;
			if (Ops[i][OLEN] > nmatch) do
				pos := %1;
				i := i-1;
			end
		end
		i := i+1;
	end
	Rejected := c;
	if (pos \= %1) do
		str.copy(Text, Ops[pos][ONAME]);
		Flags := Ops[pos][OFLGS];
	end
	Tlength := nmatch;
	return pos = %1-> 0: Ops[pos][OTOK];
end


scan() do
	var	c, i, b::3;
	var	radix;

	Flags := 0;
	c := getc();
	while (	c = '\s' \/ c = '\t' \/ c = '\n' \/
		c = '\r' \/ c = '\f'
	) do
		if (c = '\n') Line := Line+1;
		if (\O_adjust) putc(c);
		c := getc();
	end
	if (In.eof()) return ENDFILE;
	i := 0;
	if (c = '!') do
		while (c \= '\n' /\ \In.eof()) do
			if (i >= TEXTLEN-1) toolong();
			Text::i := c;
			i := i+1;
			c := getc();
		end
		Text::i := 0;
		Tlength := i;
		Rejected := c;
		Flags := LPAD|PTAB|RBRK|PCOM;
		return COMMENT;
	end
	if (	'a' <= c /\ c <= 'z' \/
		'A' <= c /\ c <= 'Z' \/
		c = '_'
	) do
		while (1) do
			ie ('A' <= c /\ c <= 'Z')
				;
			else if (\('a' <= c /\ c <= 'z' \/
				'0' <= c /\ c <= '9' \/
				c = '_')
			)
				leave;
			if (i >= TEXTLEN-1) toolong();
			Text::i := c;
			i := i+1;
			c := getc();
		end
		Text::i := 0;
		Tlength := i;
		Rejected := c;
		c := findkw(Text);
		if (c) return c;
		Flags := BLNI|LSHP;
		return SYMBOL;
	end
	if ('0' <= c /\ c <= '9' \/ c = '%') do
		radix := 10;
		if (c = '%') do
			Text::0 := '%'; i := 1;
			c := getc();
		end
		if (c = '0') do
			Text::i := '0'; i := i+1;
			c := getc();
			if (c = 'x' \/ c = 'X') do
				Text::i := c; i := i+1;
				c := getc();
				radix := 16;
			end
		end
		while (1) do
			if (i >= TEXTLEN-1) toolong();
			Text::i := c; i := i+1;
			ie (radix = 16 /\ 'A' <= c /\ c <= 'F')
				;
			else ie (radix = 16 /\ 'a' <= c /\ c <= 'f')
				;
			else ie ('0' <= c /\ c <= '9')
				;
			else
				leave;
			c := getc();
		end
		Text::(i-1) := 0;
		Tlength := i-1;
		Rejected := c;
		return NUMBER;
	end
	if (c = '\'') do
		c := getce(b);
		Text::0 := ''';
		Text::1 := c; i := 2;
		if (c = '\\') do
			Text::i := b::1; i := i+1;
		end
		Text::i := '''; i := i+1;
		Text::i := 0;
		Tlength := i;
		c := getc();
		if (c \= '\'') error("missing '\\''", 0);
		return NUMBER;
	end
	if (c = '"') do
		Text::0 := '"'; i := 1;
		while (1) do
			c := getce(b);
			if (In.eof()) error("unexpected EOF", 0);
			if (c = '"') leave;
			if (i >= TEXTLEN-2) toolong();
			Text::i := c; i := i+1;
			if (b::0 = '\\') do
				Text::i := b::1; i := i+1;
			end
		end
		Text::i := '"';
		Text::(i+1) := 0;
		Tlength := i+1;
		return STRLIT;
	end
	i := findop(c);
	if (i) return i;
	error("bad input character: 0x%2:0X", c);
end


break() do var i;
	Out.wrch('\n');
	Xsize := 0;
	for (i=0, Indent+Tmpind) do
		Out.wrch('\t');
		Xsize := Xsize + Tabsize;
	end
end


print() do var tok, tag, b::100, endcls;
	endcls := 0;
	Xsize := 0;
	while (1) do
		tok := scan();
		if (tok = ENDFILE) leave;
		if (Convtab[tok][CCASE] \= NONE) do
			ie (Convtab[tok][CCASE] = UCASE)
				upcase(Text);
			else ie (Convtab[tok][CCASE] = LCASE)
				lowcase(Text);
			else if (Convtab[tok][CCASE] = CAPITAL)
				capitalize(Text);
		end
		if (O_adjust) do
			if (Flags & PCOM) Dobrk := 0;
			if (Xsize+Tlength >= Margin-1) Dobrk := 1;
			if (Dotblbrk /\ tok \= COMMENT) Dobrk := 1;
			Dotblbrk := 0;
			if (Flags & EID) do
				Indent := Indent-1;
				if (Indent < 0) do
					Indent := 0; ! CLASS...END
					endcls := 1;
				end
			end
			if (Flags & ETID) do
				Tmpind := Tmpind-1;
				if (Tmpind < 0) Tmpind := 0;
			end
			if (	Flags & BLNI /\ (Dobrk \/ Flags & LBRK) /\
				Indent = 0 /\ Tmpind = 0 /\ \Donecmnt
			) do
				Flags := Flags | BLNK;
			end
			if (Flags & BLNK /\ \(Doneblnk|Dostick)) do
				Out.wrch('\n');
				break();
				Dopad := 0;
				Dobrk := 0;
				Donebrk := 1;
				Doneblnk := 1;
			end
			if ((Flags & LBRK \/ Dobrk) /\ \(Donebrk|Dostick)) do
				break();
				Dopad := 0;
				Dobrk := 0;
				Donebrk := 1;
			end
			if (Halfpad /\ Flags & LSHP) Dopad := 1;
			if ((Flags & LPAD \/ Dopad) /\ \Donepad /\ \Donebrk) do
				ie (Dopad>1) do
					Out.wrch('\t');
					Xsize := Xsize+(Tabsize-Xsize
						mod Tabsize);
				end
				else do
					Out.wrch('\s');
					Xsize := Xsize+1;
				end
				Dopad := 0;
				Donepad := 1;
			end
			if (endcls) do
				Out.wrch('\n');
				Donebrk := 1;
				endcls := 0;
			end
		end
		if (Convtab[tok][CPREFIX]) Out.writes(Convtab[tok][CPREFIX]);
		putstr(Text);
		if (Convtab[tok][CSUFFIX]) Out.writes(Convtab[tok][CSUFFIX]);
		if (O_adjust) do
			Xsize := Xsize+Tlength;
			if (Flags & SID) Indent := Indent+1;
			if (Flags & STID) Tmpind := Tmpind+1;
			Donepad := 0;
			Donebrk := 0;
			Doneblnk := 0;
			Dostick := 0;
			Donecmnt := 0;
			Halfpad := 0;
			if (Intbl /\ Foundrb /\ Text::0 = ',') Dotblbrk := 1;
			if (Foundlb /\ Text::0 = '[') Intbl := Intbl+1;
			if (Foundrb /\ Text::0 = ']') Intbl := Intbl-1;
			if (Intbl < 0) Intbl := 0;	! safety
			Foundrb := Text::0 = ']';
			Foundlb := Text::0 = '[';
			if (Flags & RSHP) Halfpad := 1;
			if (Flags & RPAD) Dopad := Flags&PTAB->2: 1;
			if (Flags & PCOM) Donecmnt := 1;
			if (Flags & RBRK) do
				Dopad := 0;
				Halfpad := 0;
				Dobrk := 1;
			end
			if (Flags & STCK) Dostick := 1;
		end
	end
	if (Dobrk) break();
end


print_prologue() if (Prologue) Out.writes(Prologue);


print_epilogue() if (Epilogue) Out.writes(Epilogue);


fini() do
	Out.flush();
end


usage(quit) do var b::3;
	_mwrites("Usage: txprint [-c conv] [-f file] [-o option]");
	mwrites(" [-p prefix]");
	_mwrites("\t[-r ruleset] [-s suffix] [-P prologue] [-E epilogue]");
	mwrites(" [file]");
	if (quit) halt 1;
end


longusage() do
	mwrites("");
	usage(0);
	mwrites("");
	mwrites("-E line   - append line to epilogue");
	mwrites("-P line   - append line to prologue");
	mwrites("-c rule   - specify conversion rule");
	mwrites("-f file   - load rule sets from given file");
	mwrites("-o [!]opt - set or reset (!opt) option");
	mwrites("-p rule   - specify token prefix");
	mwrites("-r set    - specify a predefined rule set");
	mwrites("-s rule   - specify token suffix");
	mwrites("");
	mwrites("See $T3XDIR/txprint.rules for example rules/options");
	mwrites("");
	halt 1;
end


optarg(buf, offp, limp, np) do
	ie (offp[0]+1 < limp[0]) do
		limp[0] := offp[0];
		return @buf::(offp[0]+1);
	end
	else do
		np[0] := np[0] + 1;
		limp[0] := t.getarg(np[0], buf, 256);
		if (limp[0] < 0) usage(1);
		offp[0] := limp[0];
		return buf;
	end
end


gettarget(target)
	ie (strequ(target, "op") \/ strequ(target, "operator"))
		return OPERATOR;
	else ie (strequ(target, "kw") \/ strequ(target, "keyword"))
		return KEYWORD;
	else ie (strequ(target, "sym") \/ strequ(target, "symbol"))
		return SYMBOL;
	else ie (strequ(target, "num") \/ strequ(target, "number"))
		return NUMBER;
	else ie (strequ(target, "str") \/ strequ(target, "string"))
		return STRLIT;
	else ie (strequ(target, "com") \/ strequ(target, "comment"))
		return COMMENT;
	else
		error("bad target: %S", target);


conv_rule(s) do
	var	target::128, type::128;
	var	tg, tp;

	if (str.length(s) > 128) error("rule too long", s);
	if (str.parse(s, "%S:%S", [(target), (type)]) \= 2)
		error("bad rule: %S", s);
	lowcase(target);
	lowcase(type);
	tg := gettarget(target);
	ie (strequ(type, "lc") \/ strequ(type, "lowercase"))
		tp := LCASE;
	else ie (strequ(type, "uc") \/ strequ(type, "uppercase"))
		tp := UCASE;
	else ie (strequ(type, "cap") \/ strequ(type, "capital"))
		tp := CAPITAL;
	else ie (strequ(type, "none") \/ strequ(type, "pass"))
		tp := NONE;
	else
		error("bad conversion type: %S", type);
	Convtab[tg][CCASE] := tp;
end


savestr(s) do var k, p;
	k := str.length(s)+1;
	p := mem.alloc(k);
	if (\p) error("out of string memory", 0);
	str.copy(p, s);
	return p;
end


append(pa, b) do var n, k1, k2;
	if (pa[0] = 0) do
		pa[0] := savestr(b);
		append(pa, "");
		return;
	end
	k1 := str.length(pa[0]);
	k2 := str.length(b)+2;
	n := mem.alloc(k1+k2);
	if (\n) error("out of string memory", 0);
	str.copy(n, pa[0]);
	str.copy(@n::k1, b);
	n::(k1+k2-2) := '\n';
	n::(k1+k2-1) := 0;
	mem.free(pa[0]);
	pa[0] := n;
end


prefix_rule(s) do
	var	target::128, pfx::128;
	var	tg;

	if (str.length(s) > 128) error("rule too long", s);
	if (str.parse(s, "%S:%S", [(target), (pfx)]) \= 2)
		error("bad rule: %S", s);
	lowcase(target);
	tg := gettarget(target);
	Convtab[tg][CPREFIX] := savestr(pfx);
end


suffix_rule(s) do
	var	target::128, sfx::128;
	var	tg;

	if (str.length(s) > 128) error("rule too long", s);
	if (str.parse(s, "%S:%S", [(target), (sfx)]) \= 2)
		error("bad rule: %S", s);
	lowcase(target);
	tg := gettarget(target);
	Convtab[tg][CSUFFIX] := savestr(sfx);
end


setopt(opt) do var flag;
	flag := 1;
	if (opt::0 = '!') do
		opt := @opt::1;
		flag := 0;
	end
	lowcase(opt);
	ie (strequ(opt, "html")) O_html := flag;
	else ie (strequ(opt, "adjust")) O_adjust := flag;
	else error("%S: unknown option", opt);
end


loadrules(id) do
	object	rules[iostream];
	var	rbuf::BUFLEN;
	var	b::TEXTLEN, k;
	var	group, arg::TEXTLEN;
	var	i, l;
	var	path::256;
	var	locations;

	ie (Rulefile) do
		if (rules.open(Rulefile, rbuf, BUFLEN,
			iostream.FREAD|iostream.FTRANS) = %1
		)
			error("%S: cannot open rule file", Rulefile);
	end
	else do
		locations := [ ".", "$HOME", "$T3XDIR" ];
		for (i=0, 3) do
			ie (locations[i]::0 = '$') do
				k := t.getenv(@locations[i]::1, arg, TEXTLEN);
				if (k < 0) loop;
			end
			else do
				str.copy(arg, locations[i]);
			end
			ie (i = 1)
				str.format(path, "%S/.txprint.rules", [(arg)]);
			else
				str.format(path, "%S/txprint.rules", [(arg)]);
			if (rules.open(path , rbuf, BUFLEN,
				iostream.FREAD|iostream.FTRANS) \= %1
			)
				leave;
		end
		if (i > 2) error(
			"no rule file found (tried . $HOME $T3XDIR)", 0);
	end
	k := rules.reads(b, TEXTLEN);
	while (\rules.eof()) do
		b::(k-1) := 0;
		if (strequ(id, b)) leave;
		k := rules.reads(b, TEXTLEN);
	end
	if (rules.eof()) do
		rules.close();
		error("%S: no such rule set", id);
	end
	l := 1;
	k := rules.reads(b, TEXTLEN);
	while (\rules.eof()) do
		b::(k-1) := 0;
		if (str.parse(b, "%W-%C%W%S", [(@group), (arg)]) \= 2)
			leave;
		ie (group = 'c') conv_rule(arg);
		else ie (group = 'o') setopt(arg);
		else ie (group = 'p') prefix_rule(arg);
		else ie (group = 's') suffix_rule(arg);
		else ie (group = 'E') append(@Epilogue, arg);
		else ie (group = 'P') append(@Prologue, arg);
		else error("%C: bad group in rule set", group);
		k := rules.reads(b, TEXTLEN);
		l := l+1;
	end
	if (\rules.eof() /\ chr.space(b::0)) do
		Line := l;
		error("%S: syntax error in rule set", id);
	end
	rules.close();
end


options() do
	var	b::256;
	var	i, j, k, c;

	i := 1;
	while (1) do
		k := t.getarg(i, b, 256);
		if (k < 0 \/ b::0 \= '-') leave;
		for (j=1, k) do
			c := b::j;
			ie (c = 'c') conv_rule(optarg(@b, @j, @k, @i));
			else ie (c = 'f') Rulefile :=
				savestr(optarg(@b, @j, @k, @i));
			else ie (c = 'o') setopt(optarg(@b, @j, @k, @i));
			else ie (c = 'r') loadrules(optarg(@b, @j, @k, @i));
			else ie (c = 'p') prefix_rule(optarg(@b, @j, @k, @i));
			else ie (c = 's') suffix_rule(optarg(@b, @j, @k, @i));
			else ie (c = 'E')
				append(@Epilogue, optarg(@b, @j, @k, @i));
			else ie (c = 'P')
				append(@Prologue, optarg(@b, @j, @k, @i));
			else ie (c = '?')
				longusage();
			else usage(1);
		end
		i := i+1;
	end
	return k<0-> 0: savestr(b);
end


do var name;
	init();
	name := options();
	if (name) do
		if (In.open(name, Inbuf, BUFLEN, iostream.FREAD) = %1)
			error("%S: cannot open file", name);
	end
	print_prologue();
	print();
	print_epilogue();
	if (name) In.close();
	fini();
end
